# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc
import warnings

from gRPC_impl.mytorch.utils.data import data_pb2 as mytorch_dot_utils_dot_data_dot_data__pb2

GRPC_GENERATED_VERSION = '1.71.0'
GRPC_VERSION = grpc.__version__
_version_not_supported = False

try:
    from grpc._utilities import first_version_is_lower
    _version_not_supported = first_version_is_lower(GRPC_VERSION, GRPC_GENERATED_VERSION)
except ImportError:
    _version_not_supported = True

if _version_not_supported:
    raise RuntimeError(
        f'The grpc package installed is at version {GRPC_VERSION},'
        + f' but the generated code in mytorch/utils/data/data_pb2_grpc.py depends on'
        + f' grpcio>={GRPC_GENERATED_VERSION}.'
        + f' Please upgrade your grpc module to grpcio>={GRPC_GENERATED_VERSION}'
        + f' or downgrade your generated code using grpcio-tools<={GRPC_VERSION}.'
    )


class DataServiceStub(object):
    """/////////////////////////////////////////////////////////////////////////////////
    Service definition
    /////////////////////////////////////////////////////////////////////////////////
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.CreateSubsetOnServer = channel.unary_unary(
                '/mytorch.DataService/CreateSubsetOnServer',
                request_serializer=mytorch_dot_utils_dot_data_dot_data__pb2.CreateSubsetRequest.SerializeToString,
                response_deserializer=mytorch_dot_utils_dot_data_dot_data__pb2.GrpcSubset.FromString,
                _registered_method=True)
        self.CreateDataLoaderOnServer = channel.unary_unary(
                '/mytorch.DataService/CreateDataLoaderOnServer',
                request_serializer=mytorch_dot_utils_dot_data_dot_data__pb2.CreateDataLoaderRequest.SerializeToString,
                response_deserializer=mytorch_dot_utils_dot_data_dot_data__pb2.GrpcDataLoader.FromString,
                _registered_method=True)
        self.GetNextDataLoaderBatch = channel.unary_unary(
                '/mytorch.DataService/GetNextDataLoaderBatch',
                request_serializer=mytorch_dot_utils_dot_data_dot_data__pb2.GetNextBatchRequest.SerializeToString,
                response_deserializer=mytorch_dot_utils_dot_data_dot_data__pb2.DataLoaderBatchResponse.FromString,
                _registered_method=True)


class DataServiceServicer(object):
    """/////////////////////////////////////////////////////////////////////////////////
    Service definition
    /////////////////////////////////////////////////////////////////////////////////
    """

    def CreateSubsetOnServer(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def CreateDataLoaderOnServer(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetNextDataLoaderBatch(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_DataServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'CreateSubsetOnServer': grpc.unary_unary_rpc_method_handler(
                    servicer.CreateSubsetOnServer,
                    request_deserializer=mytorch_dot_utils_dot_data_dot_data__pb2.CreateSubsetRequest.FromString,
                    response_serializer=mytorch_dot_utils_dot_data_dot_data__pb2.GrpcSubset.SerializeToString,
            ),
            'CreateDataLoaderOnServer': grpc.unary_unary_rpc_method_handler(
                    servicer.CreateDataLoaderOnServer,
                    request_deserializer=mytorch_dot_utils_dot_data_dot_data__pb2.CreateDataLoaderRequest.FromString,
                    response_serializer=mytorch_dot_utils_dot_data_dot_data__pb2.GrpcDataLoader.SerializeToString,
            ),
            'GetNextDataLoaderBatch': grpc.unary_unary_rpc_method_handler(
                    servicer.GetNextDataLoaderBatch,
                    request_deserializer=mytorch_dot_utils_dot_data_dot_data__pb2.GetNextBatchRequest.FromString,
                    response_serializer=mytorch_dot_utils_dot_data_dot_data__pb2.DataLoaderBatchResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'mytorch.DataService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('mytorch.DataService', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class DataService(object):
    """/////////////////////////////////////////////////////////////////////////////////
    Service definition
    /////////////////////////////////////////////////////////////////////////////////
    """

    @staticmethod
    def CreateSubsetOnServer(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/mytorch.DataService/CreateSubsetOnServer',
            mytorch_dot_utils_dot_data_dot_data__pb2.CreateSubsetRequest.SerializeToString,
            mytorch_dot_utils_dot_data_dot_data__pb2.GrpcSubset.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def CreateDataLoaderOnServer(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/mytorch.DataService/CreateDataLoaderOnServer',
            mytorch_dot_utils_dot_data_dot_data__pb2.CreateDataLoaderRequest.SerializeToString,
            mytorch_dot_utils_dot_data_dot_data__pb2.GrpcDataLoader.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetNextDataLoaderBatch(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/mytorch.DataService/GetNextDataLoaderBatch',
            mytorch_dot_utils_dot_data_dot_data__pb2.GetNextBatchRequest.SerializeToString,
            mytorch_dot_utils_dot_data_dot_data__pb2.DataLoaderBatchResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)



from gRPC_impl.mytorch.scaffolding import scaffolding_pb2, scaffolding_pb2_grpc
from gRPC_impl import shared_msg_types_pb2
from connection_utils.server_connection import ServerConnection, wrap_with_error_handler
from utils.logger import Logger

class ScaffoldingProxy:
    def __init__(self):
        self.channel = ServerConnection.get_active_connection()
        self.stub = scaffolding_pb2_grpc.ScaffoldingServiceStub(self.channel)
        self.logger = Logger.get_logger()

    @wrap_with_error_handler
    def register_client_with_server(self, client_info: dict):
        # convert the client info dictionary to a protobuf message
        key_value_pairs = [scaffolding_pb2.KeyValuePair(key=key, value=value) for key, value in client_info.items()]
        return self.stub.initialize_client_connection(scaffolding_pb2.ClientInfo(key_value_pairs=key_value_pairs))

    @wrap_with_error_handler
    def get_gpu_info(self):
        return self.stub.get_gpu_info(shared_msg_types_pb2.Empty())



###############################################################################
# Copyright (c) 2024 MyTorch Systems Inc. All rights reserved.
###############################################################################
from torch.Tensor import Tensor
from typing import List, Union
from numpy import ndarray
from PIL import Image
from proxies.mytorchvision.transforms.torchvision_functional_proxy import TorchvisionFunctionalProxy

# same as torchvision.transforms.functional.resize
def resize(img: Tensor, size: List[int]) -> Tensor:
    return TorchvisionFunctionalProxy().resize(img, size)

# same as torchvision.transforms.functional.to_tensor
def to_tensor(pic: Union[Image, ndarray]) -> Tensor:
    return TorchvisionFunctionalProxy().to_tensor(pic)


from gRPC_impl import shared_msg_types_pb2 as _shared_msg_types_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Optional as _Optional

DESCRIPTOR: _descriptor.FileDescriptor

class GrpcTransform(_message.Message):
    __slots__ = ("uuid",)
    UUID_FIELD_NUMBER: _ClassVar[int]
    uuid: str
    def __init__(self, uuid: _Optional[str] = ...) -> None: ...

class CreateCenterCropTransformRequest(_message.Message):
    __slots__ = ("size",)
    SIZE_FIELD_NUMBER: _ClassVar[int]
    size: int
    def __init__(self, size: _Optional[int] = ...) -> None: ...

class CreateResizeTransformRequest(_message.Message):
    __slots__ = ("size",)
    SIZE_FIELD_NUMBER: _ClassVar[int]
    size: int
    def __init__(self, size: _Optional[int] = ...) -> None: ...

class CreateNormalizeTransformRequest(_message.Message):
    __slots__ = ("mean", "std")
    MEAN_FIELD_NUMBER: _ClassVar[int]
    STD_FIELD_NUMBER: _ClassVar[int]
    mean: _containers.RepeatedScalarFieldContainer[float]
    std: _containers.RepeatedScalarFieldContainer[float]
    def __init__(self, mean: _Optional[_Iterable[float]] = ..., std: _Optional[_Iterable[float]] = ...) -> None: ...

class CreateComposeTransformRequest(_message.Message):
    __slots__ = ("component_transform_uuid",)
    COMPONENT_TRANSFORM_UUID_FIELD_NUMBER: _ClassVar[int]
    component_transform_uuid: _containers.RepeatedScalarFieldContainer[str]
    def __init__(self, component_transform_uuid: _Optional[_Iterable[str]] = ...) -> None: ...



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: mytorch/resnet/resnet.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1bmytorch/resnet/resnet.proto\x12\x07mytorch\"u\n\x1a\x43reateResidualBlockRequest\x12\x13\n\x0bin_channels\x18\x01 \x01(\x05\x12\x14\n\x0cout_channels\x18\x02 \x01(\x05\x12\x0e\n\x06stride\x18\x03 \x01(\x05\x12\x1c\n\x14\x64ownample_model_uuid\x18\x04 \x01(\t\"1\n\x1b\x43reateResidualBlockResponse\x12\x12\n\nmodel_uuid\x18\x01 \x01(\t\"?\n\x18\x43reateResNetModelRequest\x12\x13\n\x0bnum_classes\x18\x01 \x01(\x05\x12\x0e\n\x06layers\x18\x02 \x03(\x05\"/\n\x19\x43reateResNetModelResponse\x12\x12\n\nmodel_uuid\x18\x01 \x01(\t2\xcd\x01\n\rResnetService\x12`\n\x13\x43reateResidualBlock\x12#.mytorch.CreateResidualBlockRequest\x1a$.mytorch.CreateResidualBlockResponse\x12Z\n\x11\x43reateResNetModel\x12!.mytorch.CreateResNetModelRequest\x1a\".mytorch.CreateResNetModelResponseb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'mytorch.resnet.resnet_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_CREATERESIDUALBLOCKREQUEST']._serialized_start=40
  _globals['_CREATERESIDUALBLOCKREQUEST']._serialized_end=157
  _globals['_CREATERESIDUALBLOCKRESPONSE']._serialized_start=159
  _globals['_CREATERESIDUALBLOCKRESPONSE']._serialized_end=208
  _globals['_CREATERESNETMODELREQUEST']._serialized_start=210
  _globals['_CREATERESNETMODELREQUEST']._serialized_end=273
  _globals['_CREATERESNETMODELRESPONSE']._serialized_start=275
  _globals['_CREATERESNETMODELRESPONSE']._serialized_end=322
  _globals['_RESNETSERVICE']._serialized_start=325
  _globals['_RESNETSERVICE']._serialized_end=530
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: mytransformers/HuggingFaceModel.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from gRPC_impl import shared_msg_types_pb2 as shared__msg__types__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n%mytransformers/HuggingFaceModel.proto\x12\x0emytransformers\x1a\x16shared_msg_types.proto\"\xd5\x01\n\x16\x66romPretrained_request\x12%\n\x1dpretrained_model_name_or_path\x18\x01 \x01(\t\x12\x12\n\nmodel_args\x18\x02 \x03(\t\x12\x42\n\x06kwargs\x18\x03 \x03(\x0b\x32\x32.mytransformers.fromPretrained_request.KwargsEntry\x1a<\n\x0bKwargsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x1c\n\x05value\x18\x02 \x01(\x0b\x32\r.shared.Value:\x02\x38\x01\"-\n\x17\x66romPretrained_response\x12\x12\n\nmodel_uuid\x18\x01 \x01(\t\"L\n\x10generate_request\x12\x12\n\nmodel_uuid\x18\x01 \x01(\t\x12$\n\x06kwargs\x18\x02 \x03(\x0b\x32\x14.shared.keyValuePair\"(\n\x11generate_response\x12\x13\n\x0btensor_uuid\x18\x01 \x01(\t2\x9c\x07\n\x17HuggingFaceModelService\x12j\n\x17getAutoModelForCausalML\x12&.mytransformers.fromPretrained_request\x1a\'.mytransformers.fromPretrained_response\x12x\n%getAutoModelForSequenceClassification\x12&.mytransformers.fromPretrained_request\x1a\'.mytransformers.fromPretrained_response\x12x\n%getBertModelForSequenceClassification\x12&.mytransformers.fromPretrained_request\x1a\'.mytransformers.fromPretrained_response\x12{\n(getRobertaModelForSequenceClassification\x12&.mytransformers.fromPretrained_request\x1a\'.mytransformers.fromPretrained_response\x12p\n\x1dgetT5ForConditionalGeneration\x12&.mytransformers.fromPretrained_request\x1a\'.mytransformers.fromPretrained_response\x12k\n\x18getAutoModelForSeq2SeqLM\x12&.mytransformers.fromPretrained_request\x1a\'.mytransformers.fromPretrained_response\x12t\n!getXLNetForSequenceClassification\x12&.mytransformers.fromPretrained_request\x1a\'.mytransformers.fromPretrained_response\x12O\n\x08generate\x12 .mytransformers.generate_request\x1a!.mytransformers.generate_responseb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'mytransformers.HuggingFaceModel_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_FROMPRETRAINED_REQUEST_KWARGSENTRY']._loaded_options = None
  _globals['_FROMPRETRAINED_REQUEST_KWARGSENTRY']._serialized_options = b'8\001'
  _globals['_FROMPRETRAINED_REQUEST']._serialized_start=82
  _globals['_FROMPRETRAINED_REQUEST']._serialized_end=295
  _globals['_FROMPRETRAINED_REQUEST_KWARGSENTRY']._serialized_start=235
  _globals['_FROMPRETRAINED_REQUEST_KWARGSENTRY']._serialized_end=295
  _globals['_FROMPRETRAINED_RESPONSE']._serialized_start=297
  _globals['_FROMPRETRAINED_RESPONSE']._serialized_end=342
  _globals['_GENERATE_REQUEST']._serialized_start=344
  _globals['_GENERATE_REQUEST']._serialized_end=420
  _globals['_GENERATE_RESPONSE']._serialized_start=422
  _globals['_GENERATE_RESPONSE']._serialized_end=462
  _globals['_HUGGINGFACEMODELSERVICE']._serialized_start=465
  _globals['_HUGGINGFACEMODELSERVICE']._serialized_end=1389
# @@protoc_insertion_point(module_scope)



###############################################################################
# Copyright (c) 2024 MyTorch Systems Inc. All rights reserved.
###############################################################################
from .T5TokenizerFast import T5TokenizerFast
from proxies.mytransformers.HuggingFaceTokenizer_proxy import HuggingFaceTokenizerProxy

class T5Tokenizer(T5TokenizerFast):

    def __init__(self, uuid=None, special_tokens: dict = None):
        super().__init__(uuid, special_tokens)

    @classmethod
    def _get_instance_data(cls, pretrained_model_name_or_path, *inputs, **kwargs):
        return HuggingFaceTokenizerProxy().getT5Tokenizer(pretrained_model_name_or_path, *inputs, **kwargs)


###############################################################################
# Copyright (c) 2024 MyTorch Systems Inc. All rights reserved.
###############################################################################

from torch.utils.data.Dataset import Dataset
from proxies.mytorch.utils.data.dataloader_proxy import DataLoaderProxy

class DataLoader:
    def __init__(self, dataset: Dataset, batch_size: int, shuffle: bool) -> None:
        super().__init__()
        self.dataset = dataset
        self.shuffle = shuffle
        self.proxy = DataLoaderProxy()
        self.uuid, self.dataset_length, self.batch_size = self.createDataLoaderOnServer(dataset, batch_size, shuffle)
        self.batch_index = 0  # Initialize batch index

    def createDataLoaderOnServer(self, dataset: Dataset, batch_size: int, shuffle: bool):
        uuid, dataset_length, batch_size = self.proxy.createDataLoaderOnServer(dataset, batch_size, shuffle)
        return uuid, dataset_length, batch_size
    
    def __len__(self):
        return self.dataset_length

    def __iter__(self):
        """
        Makes the DataLoader iterable. The proxy returns a list of (input, target) tuples.
        """
        return iter(self.proxy)

    def __getitem__(self, index):
        """
        Direct indexing of batches. Returns a tuple (inputs, targets).
        """
        return self.proxy[index]

# Export everything from this module
__all__ = ['DataLoader']



from gRPC_impl import shared_msg_types_pb2 as _shared_msg_types_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Optional as _Optional

DESCRIPTOR: _descriptor.FileDescriptor

class ResizeRequest(_message.Message):
    __slots__ = ("tensor_uuid", "size")
    TENSOR_UUID_FIELD_NUMBER: _ClassVar[int]
    SIZE_FIELD_NUMBER: _ClassVar[int]
    tensor_uuid: str
    size: _containers.RepeatedScalarFieldContainer[int]
    def __init__(self, tensor_uuid: _Optional[str] = ..., size: _Optional[_Iterable[int]] = ...) -> None: ...



from gRPC_impl import shared_msg_types_pb2 as _shared_msg_types_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class fromPretrained_request(_message.Message):
    __slots__ = ("pretrained_model_name_or_path", "model_args", "kwargs")
    class KwargsEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: _shared_msg_types_pb2.Value
        def __init__(self, key: _Optional[str] = ..., value: _Optional[_Union[_shared_msg_types_pb2.Value, _Mapping]] = ...) -> None: ...
    PRETRAINED_MODEL_NAME_OR_PATH_FIELD_NUMBER: _ClassVar[int]
    MODEL_ARGS_FIELD_NUMBER: _ClassVar[int]
    KWARGS_FIELD_NUMBER: _ClassVar[int]
    pretrained_model_name_or_path: str
    model_args: _containers.RepeatedScalarFieldContainer[str]
    kwargs: _containers.MessageMap[str, _shared_msg_types_pb2.Value]
    def __init__(self, pretrained_model_name_or_path: _Optional[str] = ..., model_args: _Optional[_Iterable[str]] = ..., kwargs: _Optional[_Mapping[str, _shared_msg_types_pb2.Value]] = ...) -> None: ...

class fromPretrained_response(_message.Message):
    __slots__ = ("model_uuid",)
    MODEL_UUID_FIELD_NUMBER: _ClassVar[int]
    model_uuid: str
    def __init__(self, model_uuid: _Optional[str] = ...) -> None: ...

class generate_request(_message.Message):
    __slots__ = ("model_uuid", "kwargs")
    MODEL_UUID_FIELD_NUMBER: _ClassVar[int]
    KWARGS_FIELD_NUMBER: _ClassVar[int]
    model_uuid: str
    kwargs: _containers.RepeatedCompositeFieldContainer[_shared_msg_types_pb2.keyValuePair]
    def __init__(self, model_uuid: _Optional[str] = ..., kwargs: _Optional[_Iterable[_Union[_shared_msg_types_pb2.keyValuePair, _Mapping]]] = ...) -> None: ...

class generate_response(_message.Message):
    __slots__ = ("tensor_uuid",)
    TENSOR_UUID_FIELD_NUMBER: _ClassVar[int]
    tensor_uuid: str
    def __init__(self, tensor_uuid: _Optional[str] = ...) -> None: ...



###############################################################################
# Copyright (c) 2024 MyTorch Systems Inc. All rights reserved.
###############################################################################

'''
An AutoTokenizer is a generalized factory class that can be used to instantiate a tokenizer from a pretrained model.

For example, calling AutoTokenizer.from_pretrained(pretrained_model_name_or_path) given a BERT model name or path
is the same as calling BertTokenizer.from_pretrained(pretrained_model_name_or_path).
'''

from proxies.mytransformers.HuggingFaceTokenizer_proxy import HuggingFaceTokenizerProxy
from transformers.PreTrainedTokenizerBase import PreTrainedTokenizerBase

class AutoTokenizer(PreTrainedTokenizerBase):

    def __init__(self, uuid=None, special_tokens: dict = None):
        super().__init__(uuid, special_tokens)

    @classmethod
    def _get_instance_data(cls, pretrained_model_name_or_path, *inputs, **kwargs):
        return HuggingFaceTokenizerProxy().getAutoTokenizer(pretrained_model_name_or_path, *inputs, **kwargs)


from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class tokenizer_method_arg_pair(_message.Message):
    __slots__ = ("key", "string_value", "int_value", "float_value", "bool_value")
    KEY_FIELD_NUMBER: _ClassVar[int]
    STRING_VALUE_FIELD_NUMBER: _ClassVar[int]
    INT_VALUE_FIELD_NUMBER: _ClassVar[int]
    FLOAT_VALUE_FIELD_NUMBER: _ClassVar[int]
    BOOL_VALUE_FIELD_NUMBER: _ClassVar[int]
    key: str
    string_value: str
    int_value: int
    float_value: float
    bool_value: bool
    def __init__(self, key: _Optional[str] = ..., string_value: _Optional[str] = ..., int_value: _Optional[int] = ..., float_value: _Optional[float] = ..., bool_value: bool = ...) -> None: ...

class encode_plus_request(_message.Message):
    __slots__ = ("tokenizer_uuid", "text", "kwargs")
    TOKENIZER_UUID_FIELD_NUMBER: _ClassVar[int]
    TEXT_FIELD_NUMBER: _ClassVar[int]
    KWARGS_FIELD_NUMBER: _ClassVar[int]
    tokenizer_uuid: str
    text: str
    kwargs: _containers.RepeatedCompositeFieldContainer[tokenizer_method_arg_pair]
    def __init__(self, tokenizer_uuid: _Optional[str] = ..., text: _Optional[str] = ..., kwargs: _Optional[_Iterable[_Union[tokenizer_method_arg_pair, _Mapping]]] = ...) -> None: ...

class tokenizer_key_uuid_pair(_message.Message):
    __slots__ = ("key", "tensor_uuid")
    KEY_FIELD_NUMBER: _ClassVar[int]
    TENSOR_UUID_FIELD_NUMBER: _ClassVar[int]
    key: str
    tensor_uuid: str
    def __init__(self, key: _Optional[str] = ..., tensor_uuid: _Optional[str] = ...) -> None: ...

class encode_plus_response(_message.Message):
    __slots__ = ("encoding_uuid", "data")
    ENCODING_UUID_FIELD_NUMBER: _ClassVar[int]
    DATA_FIELD_NUMBER: _ClassVar[int]
    encoding_uuid: str
    data: _containers.RepeatedCompositeFieldContainer[tokenizer_key_uuid_pair]
    def __init__(self, encoding_uuid: _Optional[str] = ..., data: _Optional[_Iterable[_Union[tokenizer_key_uuid_pair, _Mapping]]] = ...) -> None: ...

class set_pad_token_request(_message.Message):
    __slots__ = ("tokenizer_uuid", "pad_token")
    TOKENIZER_UUID_FIELD_NUMBER: _ClassVar[int]
    PAD_TOKEN_FIELD_NUMBER: _ClassVar[int]
    tokenizer_uuid: str
    pad_token: str
    def __init__(self, tokenizer_uuid: _Optional[str] = ..., pad_token: _Optional[str] = ...) -> None: ...

class set_pad_token_response(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class decode_request(_message.Message):
    __slots__ = ("tokenizer_uuid", "input_tensor_uuid", "kwargs")
    TOKENIZER_UUID_FIELD_NUMBER: _ClassVar[int]
    INPUT_TENSOR_UUID_FIELD_NUMBER: _ClassVar[int]
    KWARGS_FIELD_NUMBER: _ClassVar[int]
    tokenizer_uuid: str
    input_tensor_uuid: str
    kwargs: _containers.RepeatedCompositeFieldContainer[tokenizer_method_arg_pair]
    def __init__(self, tokenizer_uuid: _Optional[str] = ..., input_tensor_uuid: _Optional[str] = ..., kwargs: _Optional[_Iterable[_Union[tokenizer_method_arg_pair, _Mapping]]] = ...) -> None: ...

class decode_response(_message.Message):
    __slots__ = ("text",)
    TEXT_FIELD_NUMBER: _ClassVar[int]
    text: str
    def __init__(self, text: _Optional[str] = ...) -> None: ...



###############################################################################
# Copyright (c) 2024-2025 MyTorch Systems Inc. All rights reserved.
###############################################################################
from proxies.base_proxy import BaseProxy
from torch.utils.data.Dataset import Dataset
from utils.logger import Logger
from connection_utils.server_connection import ServerConnection, wrap_with_error_handler

class TensorDatasetProxy(BaseProxy):
    def __init__(self):
        super().__init__()
        self.logger = Logger.get_logger()
        self.channel = ServerConnection.get_active_connection()

    def createTensorDatasetOnServer(self, dataset: Dataset):
        self.uuid, self.len = self.generic_call("torch.utils.data", 
                                                "TensorDataset", 
                                                dataset.uuid,
                                                call_type="constructor")
        return self.uuid, self.len
    



def test_save_load_model():
    import torch
    import hashlib
    import os

    # Create simple tensor state_dict
    a = torch.randn(3, 3)
    b = torch.ones(2, 2)

    state_dict = {
        "layer1.weight": a,
        "layer2.bias": b,
    }

    # Save using PyTorch
    torch.save( state_dict, "saved_model.pt" )

    # Load and verify, 
    loaded = torch.load("saved_model.pt")
    assert "layer1.weight" in loaded
    assert "layer2.bias" in loaded
    assert torch.equal(loaded["layer1.weight"], a)
    assert torch.equal(loaded["layer2.bias"], b)

    print("Simple model save test passed.")

def test_resnet18_save_load():
    import torch
    import torchvision.models as models
    import os

    # Load pretrained ResNet-18 and switch to eval mode
    model = models.resnet18(weights=models.ResNet18_Weights.DEFAULT)
    model.eval()

    # Extract and save the state_dict
    original_state_dict = model.state_dict()
    torch.save(original_state_dict, "resnet18_saved.pt")

    # Load state_dict into a fresh instance
    model_loaded = models.resnet18(weights=models.ResNet18_Weights.DEFAULT)
    model_loaded.load_state_dict(torch.load("resnet18_saved.pt"))
    model_loaded.eval()

    # Efficient comparison via state_dict keys
    loaded_sd = model_loaded.state_dict()

    assert original_state_dict.keys() == loaded_sd.keys(), "Mismatch in state_dict keys"

    for key in original_state_dict:
        v1 = original_state_dict[key]
        v2 = loaded_sd[key]
        assert torch.equal(v1, v2), f"Mismatch in parameter: {key}"

    print("✅ ResNet-18 save/load test passed.")


def main():
    test_save_load_model()
    test_resnet18_save_load()

if __name__ == "__main__":
    main()



from gRPC_impl import shared_msg_types_pb2 as _shared_msg_types_pb2
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Optional as _Optional

DESCRIPTOR: _descriptor.FileDescriptor

class OptimizerType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    SGD: _ClassVar[OptimizerType]
    ADAM: _ClassVar[OptimizerType]
    ADAGRAD: _ClassVar[OptimizerType]
    RMSPROP: _ClassVar[OptimizerType]
    ADAMAX: _ClassVar[OptimizerType]
    ADAMW: _ClassVar[OptimizerType]
    ADAMAXW: _ClassVar[OptimizerType]
    ADAGRADW: _ClassVar[OptimizerType]
    RMSPROPW: _ClassVar[OptimizerType]
SGD: OptimizerType
ADAM: OptimizerType
ADAGRAD: OptimizerType
RMSPROP: OptimizerType
ADAMAX: OptimizerType
ADAMW: OptimizerType
ADAMAXW: OptimizerType
ADAGRADW: OptimizerType
RMSPROPW: OptimizerType

class CreateSGDOptimizerRequest(_message.Message):
    __slots__ = ("generator_uuid", "learning_rate", "momentum")
    GENERATOR_UUID_FIELD_NUMBER: _ClassVar[int]
    LEARNING_RATE_FIELD_NUMBER: _ClassVar[int]
    MOMENTUM_FIELD_NUMBER: _ClassVar[int]
    generator_uuid: str
    learning_rate: float
    momentum: float
    def __init__(self, generator_uuid: _Optional[str] = ..., learning_rate: _Optional[float] = ..., momentum: _Optional[float] = ...) -> None: ...



###############################################################################
# Copyright (c) 2024 MyTorch Systems Inc. All rights reserved.
###############################################################################

from enum import Enum, auto
from torch.Tensor import Tensor
from proxies.base_proxy import BaseProxy
from utils.logger import Logger
from connection_utils.server_connection import ServerConnection, wrap_with_error_handler

############################################
# LayerType enum
############################################
class LayerType(Enum):
    LINEAR = auto()
    RELU = auto()
    FLATTEN = auto()
    SIGMOID = auto()
    MSELOSS = auto()
    # Add other layer types as needed

class LayerProxy (BaseProxy):
    def __init__(self, layer_type: LayerType, **params):
        self.layer_type = layer_type
        self.params = params
        self.logger = Logger.get_logger()
        self.channel = ServerConnection.get_active_connection()

    def print(self):
        print(f"Layer type: {self.layer_type.name}, Params: {self.params}")

    def describe(self):
        return f"Layer type: {self.layer_type.name}, Params: {self.params}"

class LinearProxy(LayerProxy):
    def __init__(self, in_features, out_features):
        super().__init__(LayerType.LINEAR, in_features=in_features, out_features=out_features)

class ReLUProxy(LayerProxy):
    def __init__(self):
        super().__init__(LayerType.RELU)

    def forward(self, tensor):
        """Calls ReLU on the server via generic_call."""
        uuid, shape, dtype = self.generic_call("torch.nn", 
                                               "ReLU", 
                                               tensor.uuid,
                                               call_type="forward")
        return Tensor(uuid, shape, dtype)
    
class FlattenProxy(LayerProxy):
    def __init__(self):
        super().__init__(LayerType.FLATTEN)

class SigmoidProxy(LayerProxy):
    def __init__(self):
        super().__init__(LayerType.SIGMOID)
    
    def forward(self, tensor):
        uuid, shape, dtype = self.generic_call("torch", 
                                               "sigmoid", 
                                               tensor.uuid, 
                                               call_type="function")
        return Tensor(uuid, shape, dtype)

class MSELossProxy(LayerProxy):
    def __init__(self, size_average=None, reduce=None, reduction: str = "mean"):
        super().__init__(LayerType.MSELOSS)

    def forward(self, input: Tensor, output: Tensor):
        """Calls ReLU on the server via generic_call."""
        uuid, shape, dtype = self.generic_call("torch.nn", 
                                               "MSELoss", 
                                               input.uuid, 
                                               output.uuid,
                                               call_type="forward")
        return Tensor(uuid, shape, dtype)


###############################################################################
# Copyright (c) 2024 MyTorch Systems Inc. All rights reserved.
###############################################################################

'''
This is a factory class that can be used to instantiate a model for Causal Language Modeling from a pretrained model.
"Causal" refers to the model's ability to predict the next token based only on the previous tokens, not on future tokens.
It's "causal" in the sense that each token can only be influenced by the tokens that came before it, maintaining a cause-effect relationship in the sequence.
'''

from transformers.models.auto.AutoModel import AutoModel
from proxies.mytransformers.HuggingFaceModel_proxy import HuggingFaceModelProxy

class AutoModelForCausalLM(AutoModel):
    def __init__(self, *args, **kwargs):
        # Call the parent class's __init__ method with the modified kwargs
        super().__init__(*args, **kwargs)

    @classmethod
    def _get_model_uuid(cls, pretrained_model_name_or_path, *args, **kwargs):
        return HuggingFaceModelProxy().getAutoModelForCausalML(pretrained_model_name_or_path, *args, **kwargs)



from gRPC_impl import shared_msg_types_pb2 as _shared_msg_types_pb2
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Optional as _Optional

DESCRIPTOR: _descriptor.FileDescriptor

class FileUploadRequest(_message.Message):
    __slots__ = ("filepath", "content")
    FILEPATH_FIELD_NUMBER: _ClassVar[int]
    CONTENT_FIELD_NUMBER: _ClassVar[int]
    filepath: str
    content: bytes
    def __init__(self, filepath: _Optional[str] = ..., content: _Optional[bytes] = ...) -> None: ...

class FolderUploadResponse(_message.Message):
    __slots__ = ("count",)
    COUNT_FIELD_NUMBER: _ClassVar[int]
    count: int
    def __init__(self, count: _Optional[int] = ...) -> None: ...

class RelativePath(_message.Message):
    __slots__ = ("path",)
    PATH_FIELD_NUMBER: _ClassVar[int]
    path: str
    def __init__(self, path: _Optional[str] = ...) -> None: ...

class DataDir(_message.Message):
    __slots__ = ("path",)
    PATH_FIELD_NUMBER: _ClassVar[int]
    path: str
    def __init__(self, path: _Optional[str] = ...) -> None: ...



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: mytorch/cuda/cuda.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from gRPC_impl import shared_msg_types_pb2 as shared__msg__types__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x17mytorch/cuda/cuda.proto\x12\x07mytorch\x1a\x16shared_msg_types.proto\"\x16\n\x14is_available_request\"-\n\x15is_available_response\x12\x14\n\x0cis_available\x18\x01 \x01(\x08\"\x15\n\x13\x65mpty_cache_request\"\x16\n\x14\x65mpty_cache_response\"\x1a\n\x18memory_allocated_request\"5\n\x19memory_allocated_response\x12\x18\n\x10memory_allocated\x18\x01 \x01(\x03\"\x19\n\x17memory_reserved_request\"3\n\x18memory_reserved_response\x12\x17\n\x0fmemory_reserved\x18\x01 \x01(\x03\"U\n\x1dget_device_properties_request\x12\x14\n\ndevice_str\x18\x01 \x01(\tH\x00\x12\x14\n\ndevice_int\x18\x02 \x01(\x05H\x00\x42\x08\n\x06\x64\x65vice\"\xeb\x01\n\x1eget_device_properties_response\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x14\n\x0ctotal_memory\x18\x02 \x01(\x03\x12\x18\n\x10\x61vailable_memory\x18\x03 \x01(\x03\x12\x18\n\x10\x61llocated_memory\x18\x04 \x01(\x03\x12\x17\n\x0freserved_memory\x18\x05 \x01(\x03\x12\x1c\n\x14max_memory_allocated\x18\x06 \x01(\x03\x12\x1b\n\x13max_memory_reserved\x18\x07 \x01(\x03\x12\x1d\n\x15multi_processor_count\x18\x08 \x01(\x05\x32\xfc\x03\n\x0b\x43udaService\x12M\n\x0cis_available\x12\x1d.mytorch.is_available_request\x1a\x1e.mytorch.is_available_response\x12J\n\x0b\x65mpty_cache\x12\x1c.mytorch.empty_cache_request\x1a\x1d.mytorch.empty_cache_response\x12Y\n\x10memory_allocated\x12!.mytorch.memory_allocated_request\x1a\".mytorch.memory_allocated_response\x12V\n\x0fmemory_reserved\x12 .mytorch.memory_reserved_request\x1a!.mytorch.memory_reserved_response\x12h\n\x15get_device_properties\x12&.mytorch.get_device_properties_request\x1a\'.mytorch.get_device_properties_response\x12\x35\n\x0fget_device_name\x12\r.shared.Empty\x1a\x13.shared.StringValueb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'mytorch.cuda.cuda_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_IS_AVAILABLE_REQUEST']._serialized_start=60
  _globals['_IS_AVAILABLE_REQUEST']._serialized_end=82
  _globals['_IS_AVAILABLE_RESPONSE']._serialized_start=84
  _globals['_IS_AVAILABLE_RESPONSE']._serialized_end=129
  _globals['_EMPTY_CACHE_REQUEST']._serialized_start=131
  _globals['_EMPTY_CACHE_REQUEST']._serialized_end=152
  _globals['_EMPTY_CACHE_RESPONSE']._serialized_start=154
  _globals['_EMPTY_CACHE_RESPONSE']._serialized_end=176
  _globals['_MEMORY_ALLOCATED_REQUEST']._serialized_start=178
  _globals['_MEMORY_ALLOCATED_REQUEST']._serialized_end=204
  _globals['_MEMORY_ALLOCATED_RESPONSE']._serialized_start=206
  _globals['_MEMORY_ALLOCATED_RESPONSE']._serialized_end=259
  _globals['_MEMORY_RESERVED_REQUEST']._serialized_start=261
  _globals['_MEMORY_RESERVED_REQUEST']._serialized_end=286
  _globals['_MEMORY_RESERVED_RESPONSE']._serialized_start=288
  _globals['_MEMORY_RESERVED_RESPONSE']._serialized_end=339
  _globals['_GET_DEVICE_PROPERTIES_REQUEST']._serialized_start=341
  _globals['_GET_DEVICE_PROPERTIES_REQUEST']._serialized_end=426
  _globals['_GET_DEVICE_PROPERTIES_RESPONSE']._serialized_start=429
  _globals['_GET_DEVICE_PROPERTIES_RESPONSE']._serialized_end=664
  _globals['_CUDASERVICE']._serialized_start=667
  _globals['_CUDASERVICE']._serialized_end=1175
# @@protoc_insertion_point(module_scope)



from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Optional as _Optional

DESCRIPTOR: _descriptor.FileDescriptor

class CreateResidualBlockRequest(_message.Message):
    __slots__ = ("in_channels", "out_channels", "stride", "downample_model_uuid")
    IN_CHANNELS_FIELD_NUMBER: _ClassVar[int]
    OUT_CHANNELS_FIELD_NUMBER: _ClassVar[int]
    STRIDE_FIELD_NUMBER: _ClassVar[int]
    DOWNAMPLE_MODEL_UUID_FIELD_NUMBER: _ClassVar[int]
    in_channels: int
    out_channels: int
    stride: int
    downample_model_uuid: str
    def __init__(self, in_channels: _Optional[int] = ..., out_channels: _Optional[int] = ..., stride: _Optional[int] = ..., downample_model_uuid: _Optional[str] = ...) -> None: ...

class CreateResidualBlockResponse(_message.Message):
    __slots__ = ("model_uuid",)
    MODEL_UUID_FIELD_NUMBER: _ClassVar[int]
    model_uuid: str
    def __init__(self, model_uuid: _Optional[str] = ...) -> None: ...

class CreateResNetModelRequest(_message.Message):
    __slots__ = ("num_classes", "layers")
    NUM_CLASSES_FIELD_NUMBER: _ClassVar[int]
    LAYERS_FIELD_NUMBER: _ClassVar[int]
    num_classes: int
    layers: _containers.RepeatedScalarFieldContainer[int]
    def __init__(self, num_classes: _Optional[int] = ..., layers: _Optional[_Iterable[int]] = ...) -> None: ...

class CreateResNetModelResponse(_message.Message):
    __slots__ = ("model_uuid",)
    MODEL_UUID_FIELD_NUMBER: _ClassVar[int]
    model_uuid: str
    def __init__(self, model_uuid: _Optional[str] = ...) -> None: ...



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc
import warnings

from gRPC_impl.mytransformers import HuggingFaceTokenizer_pb2 as mytransformers_dot_HuggingFaceTokenizer__pb2

GRPC_GENERATED_VERSION = '1.71.0'
GRPC_VERSION = grpc.__version__
_version_not_supported = False

try:
    from grpc._utilities import first_version_is_lower
    _version_not_supported = first_version_is_lower(GRPC_VERSION, GRPC_GENERATED_VERSION)
except ImportError:
    _version_not_supported = True

if _version_not_supported:
    raise RuntimeError(
        f'The grpc package installed is at version {GRPC_VERSION},'
        + f' but the generated code in mytransformers/HuggingFaceTokenizer_pb2_grpc.py depends on'
        + f' grpcio>={GRPC_GENERATED_VERSION}.'
        + f' Please upgrade your grpc module to grpcio>={GRPC_GENERATED_VERSION}'
        + f' or downgrade your generated code using grpcio-tools<={GRPC_VERSION}.'
    )


class HuggingFaceTokenizerServiceStub(object):
    """/////////////////////////////////////////////////////////////////////////////////
    Service definition
    /////////////////////////////////////////////////////////////////////////////////
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.getAutoTokenizer = channel.unary_unary(
                '/mytransformers.HuggingFaceTokenizerService/getAutoTokenizer',
                request_serializer=mytransformers_dot_HuggingFaceTokenizer__pb2.tokenizerFromPretrained_request.SerializeToString,
                response_deserializer=mytransformers_dot_HuggingFaceTokenizer__pb2.tokenizerFromPretrained_response.FromString,
                _registered_method=True)
        self.getBertTokenizer = channel.unary_unary(
                '/mytransformers.HuggingFaceTokenizerService/getBertTokenizer',
                request_serializer=mytransformers_dot_HuggingFaceTokenizer__pb2.tokenizerFromPretrained_request.SerializeToString,
                response_deserializer=mytransformers_dot_HuggingFaceTokenizer__pb2.tokenizerFromPretrained_response.FromString,
                _registered_method=True)
        self.getRobertaTokenizer = channel.unary_unary(
                '/mytransformers.HuggingFaceTokenizerService/getRobertaTokenizer',
                request_serializer=mytransformers_dot_HuggingFaceTokenizer__pb2.tokenizerFromPretrained_request.SerializeToString,
                response_deserializer=mytransformers_dot_HuggingFaceTokenizer__pb2.tokenizerFromPretrained_response.FromString,
                _registered_method=True)
        self.getT5Tokenizer = channel.unary_unary(
                '/mytransformers.HuggingFaceTokenizerService/getT5Tokenizer',
                request_serializer=mytransformers_dot_HuggingFaceTokenizer__pb2.tokenizerFromPretrained_request.SerializeToString,
                response_deserializer=mytransformers_dot_HuggingFaceTokenizer__pb2.tokenizerFromPretrained_response.FromString,
                _registered_method=True)
        self.getXLNetTokenizer = channel.unary_unary(
                '/mytransformers.HuggingFaceTokenizerService/getXLNetTokenizer',
                request_serializer=mytransformers_dot_HuggingFaceTokenizer__pb2.tokenizerFromPretrained_request.SerializeToString,
                response_deserializer=mytransformers_dot_HuggingFaceTokenizer__pb2.tokenizerFromPretrained_response.FromString,
                _registered_method=True)
        self.encode_plus = channel.unary_unary(
                '/mytransformers.HuggingFaceTokenizerService/encode_plus',
                request_serializer=mytransformers_dot_HuggingFaceTokenizer__pb2.encodePlus_request.SerializeToString,
                response_deserializer=mytransformers_dot_HuggingFaceTokenizer__pb2.encodePlus_response.FromString,
                _registered_method=True)
        self.set_pad_token = channel.unary_unary(
                '/mytransformers.HuggingFaceTokenizerService/set_pad_token',
                request_serializer=mytransformers_dot_HuggingFaceTokenizer__pb2.setPadToken_request.SerializeToString,
                response_deserializer=mytransformers_dot_HuggingFaceTokenizer__pb2.setPadToken_response.FromString,
                _registered_method=True)
        self.decode = channel.unary_unary(
                '/mytransformers.HuggingFaceTokenizerService/decode',
                request_serializer=mytransformers_dot_HuggingFaceTokenizer__pb2.decode_request.SerializeToString,
                response_deserializer=mytransformers_dot_HuggingFaceTokenizer__pb2.decode_response.FromString,
                _registered_method=True)


class HuggingFaceTokenizerServiceServicer(object):
    """/////////////////////////////////////////////////////////////////////////////////
    Service definition
    /////////////////////////////////////////////////////////////////////////////////
    """

    def getAutoTokenizer(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def getBertTokenizer(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def getRobertaTokenizer(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def getT5Tokenizer(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def getXLNetTokenizer(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def encode_plus(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def set_pad_token(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def decode(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_HuggingFaceTokenizerServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'getAutoTokenizer': grpc.unary_unary_rpc_method_handler(
                    servicer.getAutoTokenizer,
                    request_deserializer=mytransformers_dot_HuggingFaceTokenizer__pb2.tokenizerFromPretrained_request.FromString,
                    response_serializer=mytransformers_dot_HuggingFaceTokenizer__pb2.tokenizerFromPretrained_response.SerializeToString,
            ),
            'getBertTokenizer': grpc.unary_unary_rpc_method_handler(
                    servicer.getBertTokenizer,
                    request_deserializer=mytransformers_dot_HuggingFaceTokenizer__pb2.tokenizerFromPretrained_request.FromString,
                    response_serializer=mytransformers_dot_HuggingFaceTokenizer__pb2.tokenizerFromPretrained_response.SerializeToString,
            ),
            'getRobertaTokenizer': grpc.unary_unary_rpc_method_handler(
                    servicer.getRobertaTokenizer,
                    request_deserializer=mytransformers_dot_HuggingFaceTokenizer__pb2.tokenizerFromPretrained_request.FromString,
                    response_serializer=mytransformers_dot_HuggingFaceTokenizer__pb2.tokenizerFromPretrained_response.SerializeToString,
            ),
            'getT5Tokenizer': grpc.unary_unary_rpc_method_handler(
                    servicer.getT5Tokenizer,
                    request_deserializer=mytransformers_dot_HuggingFaceTokenizer__pb2.tokenizerFromPretrained_request.FromString,
                    response_serializer=mytransformers_dot_HuggingFaceTokenizer__pb2.tokenizerFromPretrained_response.SerializeToString,
            ),
            'getXLNetTokenizer': grpc.unary_unary_rpc_method_handler(
                    servicer.getXLNetTokenizer,
                    request_deserializer=mytransformers_dot_HuggingFaceTokenizer__pb2.tokenizerFromPretrained_request.FromString,
                    response_serializer=mytransformers_dot_HuggingFaceTokenizer__pb2.tokenizerFromPretrained_response.SerializeToString,
            ),
            'encode_plus': grpc.unary_unary_rpc_method_handler(
                    servicer.encode_plus,
                    request_deserializer=mytransformers_dot_HuggingFaceTokenizer__pb2.encodePlus_request.FromString,
                    response_serializer=mytransformers_dot_HuggingFaceTokenizer__pb2.encodePlus_response.SerializeToString,
            ),
            'set_pad_token': grpc.unary_unary_rpc_method_handler(
                    servicer.set_pad_token,
                    request_deserializer=mytransformers_dot_HuggingFaceTokenizer__pb2.setPadToken_request.FromString,
                    response_serializer=mytransformers_dot_HuggingFaceTokenizer__pb2.setPadToken_response.SerializeToString,
            ),
            'decode': grpc.unary_unary_rpc_method_handler(
                    servicer.decode,
                    request_deserializer=mytransformers_dot_HuggingFaceTokenizer__pb2.decode_request.FromString,
                    response_serializer=mytransformers_dot_HuggingFaceTokenizer__pb2.decode_response.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'mytransformers.HuggingFaceTokenizerService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('mytransformers.HuggingFaceTokenizerService', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class HuggingFaceTokenizerService(object):
    """/////////////////////////////////////////////////////////////////////////////////
    Service definition
    /////////////////////////////////////////////////////////////////////////////////
    """

    @staticmethod
    def getAutoTokenizer(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/mytransformers.HuggingFaceTokenizerService/getAutoTokenizer',
            mytransformers_dot_HuggingFaceTokenizer__pb2.tokenizerFromPretrained_request.SerializeToString,
            mytransformers_dot_HuggingFaceTokenizer__pb2.tokenizerFromPretrained_response.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def getBertTokenizer(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/mytransformers.HuggingFaceTokenizerService/getBertTokenizer',
            mytransformers_dot_HuggingFaceTokenizer__pb2.tokenizerFromPretrained_request.SerializeToString,
            mytransformers_dot_HuggingFaceTokenizer__pb2.tokenizerFromPretrained_response.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def getRobertaTokenizer(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/mytransformers.HuggingFaceTokenizerService/getRobertaTokenizer',
            mytransformers_dot_HuggingFaceTokenizer__pb2.tokenizerFromPretrained_request.SerializeToString,
            mytransformers_dot_HuggingFaceTokenizer__pb2.tokenizerFromPretrained_response.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def getT5Tokenizer(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/mytransformers.HuggingFaceTokenizerService/getT5Tokenizer',
            mytransformers_dot_HuggingFaceTokenizer__pb2.tokenizerFromPretrained_request.SerializeToString,
            mytransformers_dot_HuggingFaceTokenizer__pb2.tokenizerFromPretrained_response.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def getXLNetTokenizer(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/mytransformers.HuggingFaceTokenizerService/getXLNetTokenizer',
            mytransformers_dot_HuggingFaceTokenizer__pb2.tokenizerFromPretrained_request.SerializeToString,
            mytransformers_dot_HuggingFaceTokenizer__pb2.tokenizerFromPretrained_response.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def encode_plus(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/mytransformers.HuggingFaceTokenizerService/encode_plus',
            mytransformers_dot_HuggingFaceTokenizer__pb2.encodePlus_request.SerializeToString,
            mytransformers_dot_HuggingFaceTokenizer__pb2.encodePlus_response.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def set_pad_token(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/mytransformers.HuggingFaceTokenizerService/set_pad_token',
            mytransformers_dot_HuggingFaceTokenizer__pb2.setPadToken_request.SerializeToString,
            mytransformers_dot_HuggingFaceTokenizer__pb2.setPadToken_response.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def decode(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/mytransformers.HuggingFaceTokenizerService/decode',
            mytransformers_dot_HuggingFaceTokenizer__pb2.decode_request.SerializeToString,
            mytransformers_dot_HuggingFaceTokenizer__pb2.decode_response.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)



from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Optional as _Optional

DESCRIPTOR: _descriptor.FileDescriptor

class GrpcImageFolder(_message.Message):
    __slots__ = ("uuid", "image_uuids", "labels", "classes")
    UUID_FIELD_NUMBER: _ClassVar[int]
    IMAGE_UUIDS_FIELD_NUMBER: _ClassVar[int]
    LABELS_FIELD_NUMBER: _ClassVar[int]
    CLASSES_FIELD_NUMBER: _ClassVar[int]
    uuid: str
    image_uuids: _containers.RepeatedScalarFieldContainer[str]
    labels: _containers.RepeatedScalarFieldContainer[int]
    classes: _containers.RepeatedScalarFieldContainer[str]
    def __init__(self, uuid: _Optional[str] = ..., image_uuids: _Optional[_Iterable[str]] = ..., labels: _Optional[_Iterable[int]] = ..., classes: _Optional[_Iterable[str]] = ...) -> None: ...

class GrpcCIFAR10(_message.Message):
    __slots__ = ("uuid", "dataset_length", "classes")
    UUID_FIELD_NUMBER: _ClassVar[int]
    DATASET_LENGTH_FIELD_NUMBER: _ClassVar[int]
    CLASSES_FIELD_NUMBER: _ClassVar[int]
    uuid: str
    dataset_length: int
    classes: _containers.RepeatedScalarFieldContainer[str]
    def __init__(self, uuid: _Optional[str] = ..., dataset_length: _Optional[int] = ..., classes: _Optional[_Iterable[str]] = ...) -> None: ...

class CreateImageFolderRequest(_message.Message):
    __slots__ = ("relative_datapath", "transform_uuid")
    RELATIVE_DATAPATH_FIELD_NUMBER: _ClassVar[int]
    TRANSFORM_UUID_FIELD_NUMBER: _ClassVar[int]
    relative_datapath: str
    transform_uuid: str
    def __init__(self, relative_datapath: _Optional[str] = ..., transform_uuid: _Optional[str] = ...) -> None: ...

class CreateCIFAR10Request(_message.Message):
    __slots__ = ("root", "train", "download", "transform_uuid")
    ROOT_FIELD_NUMBER: _ClassVar[int]
    TRAIN_FIELD_NUMBER: _ClassVar[int]
    DOWNLOAD_FIELD_NUMBER: _ClassVar[int]
    TRANSFORM_UUID_FIELD_NUMBER: _ClassVar[int]
    root: str
    train: bool
    download: bool
    transform_uuid: str
    def __init__(self, root: _Optional[str] = ..., train: bool = ..., download: bool = ..., transform_uuid: _Optional[str] = ...) -> None: ...



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: mytorchvision/models/models.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from gRPC_impl import shared_msg_types_pb2 as shared__msg__types__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n!mytorchvision/models/models.proto\x12\rmytorchvision\x1a\x16shared_msg_types.proto\"!\n\x10WeightEnumString\x12\r\n\x05value\x18\x01 \x01(\t2\xc1\x01\n\rModelsService\x12\x39\n\x08resnet18\x12\x1f.mytorchvision.WeightEnumString\x1a\x0c.shared.UUID\x12\x39\n\x08resnet50\x12\x1f.mytorchvision.WeightEnumString\x1a\x0c.shared.UUID\x12:\n\tresnet152\x12\x1f.mytorchvision.WeightEnumString\x1a\x0c.shared.UUIDb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'mytorchvision.models.models_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_WEIGHTENUMSTRING']._serialized_start=76
  _globals['_WEIGHTENUMSTRING']._serialized_end=109
  _globals['_MODELSSERVICE']._serialized_start=112
  _globals['_MODELSSERVICE']._serialized_end=305
# @@protoc_insertion_point(module_scope)



###############################################################################
# Copyright (c) 2024 MyTorch Systems Inc. All rights reserved.
###############################################################################
from .PreTrainedTokenizerFast import PreTrainedTokenizerFast

class XLNetTokenizerFast(PreTrainedTokenizerFast):

    def __init__(self, uuid=None, special_tokens: dict = None):
        super().__init__(uuid, special_tokens)



###############################################################################
# Copyright (c) 2024 MyTorch Systems Inc. All rights reserved.
###############################################################################

"""
This singleton class gets the gRPC connection to the server.
"""
import grpc
import time

from utils.logger import Logger
from utils.MyTorchClient import MyTorchClient
from proxies.mytorch.scaffolding.resource_mgmt_proxy import ResourceMgmtProxy

class ResourceMgrConnection:

    def __init__(self, ip, port=50051):
        self.logger = Logger.get_logger()
        self.grpc_channel = None
        self.resource_mgr_ip = ip
        self.resource_mgr_port = port
        self.grpc_channel = None
        self.client_ip = None

        # Create a channel with the Resource Manager
        initial_channel = grpc.insecure_channel(self.resource_mgr_ip + ":" + str(self.resource_mgr_port))
        logging_interceptor = LoggingInterceptor()

        # Add the interceptors to the channel
        self.grpc_channel = grpc.intercept_channel(initial_channel, logging_interceptor)
        self.logger.info(f"Connected to Resource Manager {self.resource_mgr_ip}:{self.resource_mgr_port}")

    def get_destination_server_for_client(self, client_ip) -> dict:
        self.client_ip = client_ip
        auth_token = "FFFFFFFFFFFF"
        max_time = "1000"
        resource_mgr_proxy = ResourceMgmtProxy(self.grpc_channel)
        resource_info_dict = resource_mgr_proxy.resource_request(auth_token=auth_token, client_ip=self.client_ip, server_id=self.resource_mgr_ip, max_time=int(max_time))
        while resource_info_dict['resource_id'] == "":
            resource_info_dict = resource_mgr_proxy.resource_request(auth_token=auth_token, client_ip=self.client_ip, server_id=self.resource_mgr_ip, max_time=int(max_time))
            print(f"Waiting...   ", resource_info_dict['resource_id'])
            if resource_info_dict['resource_id'] == "":
                time.sleep(5)
        return resource_info_dict


###########################################################################
#### Interceptor to add client-side logging of gRPC calls
###########################################################################
class LoggingInterceptor(grpc.UnaryUnaryClientInterceptor,
                         grpc.UnaryStreamClientInterceptor,
                         grpc.StreamUnaryClientInterceptor,
                         grpc.StreamStreamClientInterceptor):
    def __init__(self):
        self.logger = Logger.get_logger()

    def intercept_unary_unary(self, continuation, client_call_details, request):
        return self._intercept_call(continuation, client_call_details, request)

    def intercept_unary_stream(self, continuation, client_call_details, request):
        return self._intercept_call(continuation, client_call_details, request)

    def intercept_stream_unary(self, continuation, client_call_details, request_iterator):
        return self._intercept_stream_call(continuation, client_call_details, request_iterator)

    def intercept_stream_stream(self, continuation, client_call_details, request_iterator):
        return self._intercept_stream_call(continuation, client_call_details, request_iterator)

    def _intercept_call(self, continuation, client_call_details, request):
        method_name = client_call_details.method
        self.logger.debug(f"~~~~~~~~~ Outgoing request ~~~~~~~~~")
        self.logger.debug(f"Calling gRPC method: {method_name}")

        start_time = time.perf_counter()
        response = continuation(client_call_details, request)
        self._log_duration(method_name, start_time)
        self._log_response(method_name, response)
        return response

    def _intercept_stream_call(self, continuation, client_call_details, request_iterator):
        method_name = client_call_details.method
        self.logger.debug(f"~~~~~~~~~ Outgoing stream request ~~~~~~~~~")
        self.logger.debug(f"Calling gRPC streaming method: {method_name}")

        start_time = time.perf_counter()

        # Wrap the request iterator to add logging
        wrapped_iterator = self._log_stream_requests(request_iterator)

        response = continuation(client_call_details, wrapped_iterator)
        self._log_duration(method_name, start_time)
        self._log_response(method_name, response)
        return response

    def _log_duration(self, method_name, start_time):
        end_time = time.perf_counter()
        duration = end_time - start_time
        rounded_duration = round(1000 * duration, 2)  # Convert to milliseconds and round to 2 decimal places
        self.logger.debug(f"Method {method_name} took: {rounded_duration} ms")
        MyTorchClient.instance().add_method_duration(method_name, duration)
        return duration

    def _log_response(self, method_name, response):
        self.logger.debug("~~~~~~~~~ Incoming response ~~~~~~~~~")
        self.logger.debug(f"Received gRPC method: {method_name}")
        self.logger.debug("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~")
        return response

    def _log_stream_requests(self, request_iterator):
        for request in request_iterator:
            self.logger.debug(f"Streaming request: {request}")
            yield request




###############################################################################
# Copyright (c) 2024 MyTorch Systems Inc. All rights reserved.
###############################################################################
from proxies.mytransformers.HuggingFaceTokenizer_proxy import HuggingFaceTokenizerProxy
from transformers.BatchEncoding import BatchEncoding
from utils.logger import Logger
from utils.spinner import Spinner

class PreTrainedTokenizerBase:
    def __init__(self, uuid=None, special_tokens: dict = None):
        self.uuid = uuid
        self.special_tokens = special_tokens or {}

        # List of special tokens to set as attributes
        token_attributes = [
            'eos_token', 'pad_token', 'bos_token', 'unk_token',
            'sep_token', 'cls_token', 'mask_token'
        ]

        # Set attributes from special_tokens dict
        # e.g. self.eos_token = special_tokens.get('eos_token')
        for attr in token_attributes:
            setattr(self, attr, self.special_tokens.get(attr))

    @classmethod
    def from_pretrained(cls, pretrained_model_name_or_path, *inputs, **kwargs):
        Logger.get_logger().info(f"Loading {cls.__name__} for `{pretrained_model_name_or_path}`")
        spinner = Spinner()
        spinner.start()

        try:
            # This method should be implemented by subclasses
            uuid, special_tokens = cls._get_instance_data(pretrained_model_name_or_path, *inputs, **kwargs)
        finally:
            spinner.stop()

        instance = cls(uuid=uuid, special_tokens=special_tokens)
        Logger.get_logger().info(f"...{cls.__name__} has been loaded")
        return instance

    @classmethod
    def _get_instance_data(cls, pretrained_model_name_or_path, *inputs, **kwargs):
        # This method should be implemented by subclasses
        raise NotImplementedError("Subclasses must implement the _get_instance_data method.")

    # Calling tokenizer() is equivalent to calling tokenizer.encode_plus()
    def __call__(self, text, **kwargs):
        return self.encode_plus(text, **kwargs)

    def encode_plus(self, text, **kwargs):
        uuid, data = HuggingFaceTokenizerProxy().encode_plus(self.uuid, text, **kwargs)
        return BatchEncoding(uuid, data)

    def decode(self, input_tensor, **kwargs):
        return HuggingFaceTokenizerProxy().decode(self.uuid, input_tensor.uuid, **kwargs)

    @property
    def pad_token(self):
        return self.pad_token

    # when the value of pad_token is set, we need to update the value on the server
    @pad_token.setter
    def pad_token(self, value):
        HuggingFaceTokenizerProxy().set_pad_token(self.uuid, value)



from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional

DESCRIPTOR: _descriptor.FileDescriptor

class tokenizer_from_pretrained_request(_message.Message):
    __slots__ = ("pretrained_model_name_or_path", "inputs", "kwargs")
    class KwargsEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
    PRETRAINED_MODEL_NAME_OR_PATH_FIELD_NUMBER: _ClassVar[int]
    INPUTS_FIELD_NUMBER: _ClassVar[int]
    KWARGS_FIELD_NUMBER: _ClassVar[int]
    pretrained_model_name_or_path: str
    inputs: _containers.RepeatedScalarFieldContainer[str]
    kwargs: _containers.ScalarMap[str, str]
    def __init__(self, pretrained_model_name_or_path: _Optional[str] = ..., inputs: _Optional[_Iterable[str]] = ..., kwargs: _Optional[_Mapping[str, str]] = ...) -> None: ...

class tokenizer_from_pretrained_response(_message.Message):
    __slots__ = ("tokenizer_uuid", "eos_token")
    TOKENIZER_UUID_FIELD_NUMBER: _ClassVar[int]
    EOS_TOKEN_FIELD_NUMBER: _ClassVar[int]
    tokenizer_uuid: str
    eos_token: str
    def __init__(self, tokenizer_uuid: _Optional[str] = ..., eos_token: _Optional[str] = ...) -> None: ...



from proxies.mytorch.nn.module_proxy import ModuleProxy
from torch.cqbase import CQBase
    
class  Module(CQBase):
    def __init__(self, uuid=None):
        CQBase.__init__(self)
        self.proxy = ModuleProxy()
        if uuid is None:
            uuid = self.proxy.create_module_on_server()
        self.set_uuid(uuid)

    def set_uuid(self, uuid):
        self.uuid = uuid
        self.proxy.uuid = uuid

    
    # It is probably possible to make a generic fix for the _in_super_call logic.
    # Also this hasn't been tested for sub-sub-classes of Module / Sequential...
    def forward(self, input_data):
        # Save the original value of _in_super_call
        original_in_super_call = getattr(self.local_data, '_in_super_call', False)
        # Set _in_super_call to True to indicate that the method is being called from a super() call
        self.local_data._in_super_call = True
        try:
            # Call the forward method on the server using the proxy
            return self.proxy.forward(input_data)
        finally:
            # Restore the original value of _in_super_call
            self.local_data._in_super_call = original_in_super_call

    def __call__(self, input_data):
        return self.forward(input_data)

    def state_dict(self) -> dict:
        return self.proxy.state_dict()

    def load_state_dict(self, dict):
        return self.proxy.load_state_dict(dict)

    def eval(self):
        self.proxy.eval()

    def to(self, device):
        self.proxy.to_device(device)

    def parameters(self):
        return self.proxy.parameters()

    def half(self):
        return self.proxy.half()

    def cuda(self):
        return self.proxy.to_device("cuda")

    def cpu(self):
        return self.proxy.to_device("cpu")
    
    def train(self, mode=True):
        self.proxy.train(mode)



###############################################################################
# Copyright (c) 2024 MyTorch Systems Inc. All rights reserved.
###############################################################################
from torchvision.models.weight_enums import WeightsEnum
from proxies.mytorchvision.models.ModelsProxy import ModelsProxy
from torch.nn.Module import Module

# example call: torchvision.models.resnet18(weights=ResNet18_Weights.DEFAULT)
def resnet18(weights: WeightsEnum) -> Module:
    model_uuid =  ModelsProxy().resnet18(weights)
    module = Module(uuid=model_uuid)
    return module

def resnet50(weights: WeightsEnum) -> Module:
    model_uuid =  ModelsProxy().resnet50(weights)
    module = Module(uuid=model_uuid)
    return module

def resnet152(weights: WeightsEnum) -> Module:
    model_uuid =  ModelsProxy().resnet152(weights)
    module = Module(uuid=model_uuid)
    return module


###############################################################################
# Copyright (c) 2024 MyTorch Systems Inc. All rights reserved.
###############################################################################

import socket
import random
import requests
from utils.logger import Logger
from connection_utils.resource_mgr_connection import ResourceMgrConnection
import torch.scaffolding as scaffolding

IP_LOOKUP_SERVICE = 'https://api.ipify.org'

class ServerResolver:

    def __init__(self, dns_name: str):
        self.dns_name = dns_name
        self.logger = Logger.get_logger()
        self.resource_mgr_ip = None
        self.client_ip = None


    '''
    This gets the destination server for MyTorch to call. It uses a DNS name
    as an input to talk to the Resource Manager server. The Resource Manager server
    will return the server that the client should talk to.
    
    There are 3 steps involved
    1. Resolve the DNS name of the Resource Manager to an IP address
    2. Get the public-facing IP of this machine
    3. Call the Resource Manager to get the server to talk to
    '''
    def connect_to_server(self) -> tuple[str, str, int] | None: # (server_name, server_ip, server_port)
        # Resolve the DNS name of the Resource Manager to an IP address and connect to it
        self.resource_mgr_ip = self.get_resource_mgr_ip()
        self.logger.info(f"Resource Manager IP address: {self.resource_mgr_ip}")
        resource_mgr_conn = ResourceMgrConnection(self.resource_mgr_ip)

        # Get the public-facing IP of this machine
        self.client_ip = self.get_public_ip()
        self.logger.info(f"Client public-facing IP address: {self.client_ip}")

        # Call the Resource Manager to get the server to talk to
        resource_info_dict = resource_mgr_conn.get_destination_server_for_client(self.client_ip)
        destination_server_name = resource_info_dict['resource_name']
        destination_server_ip = resource_info_dict['resource_ip']
        destination_server_port = resource_info_dict['resource_port']
        self.logger.info(f"Destination server: {destination_server_name} at {destination_server_ip}:{destination_server_port}")

        # connect to the destination server
        return scaffolding.select_server_by_ip(destination_server_name, destination_server_ip, destination_server_port)

    '''
    This function returns the IP address of the resource manager server
    Since the resource manager server is a DNS name, we need to resolve it to an IP address. This
    resolution will return a list of IP addresses, but we only need one. We will return a random
    IP address from the list.
    '''
    def get_resource_mgr_ip(self) -> str | None:
        try:
            # Get all IP addresses for the given DNS name
            ip_addresses = socket.getaddrinfo(self.dns_name, None)
            # Filter out only IPv4 addresses
            ipv4_addresses = [ip[4][0] for ip in ip_addresses if ip[0] == socket.AF_INET]
            # Return a random IP address from the list
            return random.choice(ipv4_addresses)
        except socket.gaierror:
            raise Exception("Could not resolve the IP address of the Resource Manager server")

    def get_public_ip(self) -> str:
        """
        Fetches the public IP address of the system.
        Returns the public IP address as a string.
        """
        try:
            # Request the public IP from a third-party service
            ip_request = requests.get(IP_LOOKUP_SERVICE)
            # Check if the request was successful
            ip_request.raise_for_status()
            # Get the public IP address from the response
            public_ip = ip_request.text
            return public_ip
        except requests.exceptions.RequestException as e:
            raise Exception(f"Error fetching public IP of this machine: {e}")


# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: mytorch/mytorch.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from gRPC_impl import shared_msg_types_pb2 as shared__msg__types__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x15mytorch/mytorch.proto\x12\x07mytorch\x1a\x16shared_msg_types.proto\"9\n\rARangeRequest\x12\r\n\x05start\x18\x01 \x01(\x02\x12\x0b\n\x03\x65nd\x18\x02 \x01(\x02\x12\x0c\n\x04step\x18\x03 \x01(\x02\"B\n\rArgMaxRequest\x12\x13\n\x0btensor_uuid\x18\x01 \x01(\t\x12\x0b\n\x03\x64im\x18\x02 \x01(\x05\x12\x0f\n\x07keepdim\x18\x03 \x01(\x08\"l\n\x0f\x41llCloseRequest\x12\x14\n\x0ctensor1_uuid\x18\x01 \x01(\t\x12\x14\n\x0ctensor2_uuid\x18\x02 \x01(\t\x12\x0c\n\x04rtol\x18\x03 \x01(\x02\x12\x0c\n\x04\x61tol\x18\x04 \x01(\x02\x12\x11\n\tequal_nan\x18\x05 \x01(\x08\"L\n\x17RepeatInterleaveRequest\x12\x13\n\x0btensor_uuid\x18\x01 \x01(\t\x12\x0f\n\x07repeats\x18\x02 \x01(\x05\x12\x0b\n\x03\x64im\x18\x03 \x01(\x05\"\xaa\x01\n\x10SaveModelRequest\x12\x10\n\x08\x66ilename\x18\x01 \x01(\t\x12K\n\x12named_tensor_uuids\x18\x02 \x03(\x0b\x32/.mytorch.SaveModelRequest.NamedTensorUuidsEntry\x1a\x37\n\x15NamedTensorUuidsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"\x91\x01\n\x11LoadModelResponse\x12\x38\n\x07tensors\x18\x01 \x03(\x0b\x32\'.mytorch.LoadModelResponse.TensorsEntry\x1a\x42\n\x0cTensorsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12!\n\x05value\x18\x02 \x01(\x0b\x32\x12.shared.GrpcTensor:\x02\x38\x01\x32\xb2\t\n\x0eMyTorchService\x12\x30\n\x05randn\x12\x13.shared.TensorShape\x1a\x12.shared.GrpcTensor\x12/\n\x04rand\x12\x13.shared.TensorShape\x1a\x12.shared.GrpcTensor\x12>\n\nfrom_numpy\x12\x1c.shared.SerializedNumpyArray\x1a\x12.shared.GrpcTensor\x12\x34\n\x06\x61range\x12\x16.mytorch.ARangeRequest\x1a\x12.shared.GrpcTensor\x12\x35\n\x07reshape\x12\x16.shared.ReshapeRequest\x1a\x12.shared.GrpcTensor\x12\x32\n\x03\x63\x61t\x12\x17.shared.TensorIDsAndDim\x1a\x12.shared.GrpcTensor\x12\x34\n\x06\x61rgmax\x12\x16.mytorch.ArgMaxRequest\x1a\x12.shared.GrpcTensor\x12\x32\n\x06matmul\x12\x14.shared.TwoTensorIDs\x1a\x12.shared.GrpcTensor\x12G\n\x11sendPytorchTensor\x12\x1e.shared.SerializedTensorObject\x1a\x12.shared.GrpcTensor\x12\x37\n\tunsqueeze\x12\x16.shared.TensorIDAndDim\x1a\x12.shared.GrpcTensor\x12\x35\n\x03max\x12\x16.shared.TensorIDAndDim\x1a\x16.shared.TwoGrpcTensors\x12@\n\x08meshgrid\x12\x17.shared.TensorIDsAndDim\x1a\x1b.shared.MultipleGrpcTensors\x12I\n\x11repeat_interleave\x12 .mytorch.RepeatInterleaveRequest\x1a\x12.shared.GrpcTensor\x12:\n\x08\x61llclose\x12\x18.mytorch.AllCloseRequest\x1a\x14.shared.BooleanValue\x12P\n$mytorch_server_get_timing_statistics\x12\x13.shared.StringValue\x1a\x13.shared.StringValue\x12X\n+mytorch_server_initialize_timing_statistics\x12\x13.shared.StringValue\x1a\x14.shared.BooleanValue\x12\x43\n\x11save_model_stream\x12\x19.mytorch.SaveModelRequest\x1a\x11.shared.FileChunk0\x01\x12\x44\n\x11load_model_stream\x12\x11.shared.FileChunk\x1a\x1a.mytorch.LoadModelResponse(\x01\x12\x39\n\x0cgeneric_call\x12\x13.shared.JsonRequest\x1a\x14.shared.JsonResponseb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'mytorch.mytorch_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_SAVEMODELREQUEST_NAMEDTENSORUUIDSENTRY']._loaded_options = None
  _globals['_SAVEMODELREQUEST_NAMEDTENSORUUIDSENTRY']._serialized_options = b'8\001'
  _globals['_LOADMODELRESPONSE_TENSORSENTRY']._loaded_options = None
  _globals['_LOADMODELRESPONSE_TENSORSENTRY']._serialized_options = b'8\001'
  _globals['_ARANGEREQUEST']._serialized_start=58
  _globals['_ARANGEREQUEST']._serialized_end=115
  _globals['_ARGMAXREQUEST']._serialized_start=117
  _globals['_ARGMAXREQUEST']._serialized_end=183
  _globals['_ALLCLOSEREQUEST']._serialized_start=185
  _globals['_ALLCLOSEREQUEST']._serialized_end=293
  _globals['_REPEATINTERLEAVEREQUEST']._serialized_start=295
  _globals['_REPEATINTERLEAVEREQUEST']._serialized_end=371
  _globals['_SAVEMODELREQUEST']._serialized_start=374
  _globals['_SAVEMODELREQUEST']._serialized_end=544
  _globals['_SAVEMODELREQUEST_NAMEDTENSORUUIDSENTRY']._serialized_start=489
  _globals['_SAVEMODELREQUEST_NAMEDTENSORUUIDSENTRY']._serialized_end=544
  _globals['_LOADMODELRESPONSE']._serialized_start=547
  _globals['_LOADMODELRESPONSE']._serialized_end=692
  _globals['_LOADMODELRESPONSE_TENSORSENTRY']._serialized_start=626
  _globals['_LOADMODELRESPONSE_TENSORSENTRY']._serialized_end=692
  _globals['_MYTORCHSERVICE']._serialized_start=695
  _globals['_MYTORCHSERVICE']._serialized_end=1897
# @@protoc_insertion_point(module_scope)



###############################################################################
# Copyright (c) 2024 MyTorch Systems Inc. All rights reserved.
###############################################################################
from proxies.mytorch.jit.jit_proxy import JitProxy
from torch.nn.Module import Module
from torch.Tensor import Tensor

def trace(model: Module, example_tensor: Tensor) -> Module:
    model_uuid = JitProxy().trace(model, example_tensor)
    module = Module(uuid=model_uuid)
    return module

def save(module: Module, path: str) -> None:
    JitProxy().save(module, path)

def load(path: str) -> Module:
    model_uuid = JitProxy().load(path)
    module = Module(uuid=model_uuid)
    return module


import threading
import itertools
import time
import sys

"""
This class provides a simple spinner that can be used to indicate that a process is running.
"""

class Spinner:
    def __init__(self, delay=0.1):
        """ Spinner constructor with delay between updates. """
        self.spinner = itertools.cycle(['-', '/', '|', '\\'])
        self.delay = delay
        self.running = False

    def spin(self):
        """ Spin the spinner by refreshing the whole line. """
        while self.running:
            sys.stdout.write('\r' + next(self.spinner) + ' Loading... ')  # Refresh the entire line
            sys.stdout.flush()
            time.sleep(self.delay)

    def start(self):
        """ Start the spinner in a separate thread. """
        self.running = True
        self.thread = threading.Thread(target=self.spin)
        self.thread.start()

    def stop(self):
        """ Stop the spinner and clean the line. """
        self.running = False
        self.thread.join()  # Wait for the spinner thread to finish
        sys.stdout.write('\r' + ' ' * 20 + '\r')  # Clear the line by overwriting with blanks
        sys.stdout.flush()



###############################################################################
# Copyright (c) 2024 MyTorch Systems Inc. All rights reserved.
###############################################################################

"""
This proxy provides a proxy for the MyTorchService gRPC service.
It allows the client to call specified torch operations on the server via gRPC.
"""

from connection_utils.server_connection import ServerConnection, wrap_with_error_handler
from gRPC_impl.mytorch_neuronx import mytorch_neuronx_pb2_grpc, mytorch_neuronx_pb2
from utils.logger import Logger
from torch.Tensor import Tensor
from torch.nn.Module import Module

class CqtorchNeuronxProxy:

    def __init__(self):
        self.channel = ServerConnection.get_active_connection()
        self.stub = mytorch_neuronx_pb2_grpc.CqtorchNeuronxServiceStub(self.channel)
        self.logger = Logger.get_logger()

    @wrap_with_error_handler
    def trace(self, model: Module, example_inputs: tuple[Tensor, ...]) -> str:
        request = mytorch_neuronx_pb2.TraceRequest() # for type hinting later on
        # get the uuid of each tensor in example_inputs
        tensor_uuid_list = [tensor.uuid for tensor in example_inputs]
        request.model_uuid = model.uuid
        request.tensor_uuid_tuple.extend(tensor_uuid_list)
        response: mytorch_neuronx_pb2.TraceResponse = self.stub.trace(request)
        return response.model_uuid



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc
import warnings

from gRPC_impl.mytransformers import HuggingFaceModel_pb2 as mytransformers_dot_HuggingFaceModel__pb2

GRPC_GENERATED_VERSION = '1.71.0'
GRPC_VERSION = grpc.__version__
_version_not_supported = False

try:
    from grpc._utilities import first_version_is_lower
    _version_not_supported = first_version_is_lower(GRPC_VERSION, GRPC_GENERATED_VERSION)
except ImportError:
    _version_not_supported = True

if _version_not_supported:
    raise RuntimeError(
        f'The grpc package installed is at version {GRPC_VERSION},'
        + f' but the generated code in mytransformers/HuggingFaceModel_pb2_grpc.py depends on'
        + f' grpcio>={GRPC_GENERATED_VERSION}.'
        + f' Please upgrade your grpc module to grpcio>={GRPC_GENERATED_VERSION}'
        + f' or downgrade your generated code using grpcio-tools<={GRPC_VERSION}.'
    )


class HuggingFaceModelServiceStub(object):
    """/////////////////////////////////////////////////////////////////////////////////
    Service definition
    /////////////////////////////////////////////////////////////////////////////////
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.getAutoModelForCausalML = channel.unary_unary(
                '/mytransformers.HuggingFaceModelService/getAutoModelForCausalML',
                request_serializer=mytransformers_dot_HuggingFaceModel__pb2.fromPretrained_request.SerializeToString,
                response_deserializer=mytransformers_dot_HuggingFaceModel__pb2.fromPretrained_response.FromString,
                _registered_method=True)
        self.getAutoModelForSequenceClassification = channel.unary_unary(
                '/mytransformers.HuggingFaceModelService/getAutoModelForSequenceClassification',
                request_serializer=mytransformers_dot_HuggingFaceModel__pb2.fromPretrained_request.SerializeToString,
                response_deserializer=mytransformers_dot_HuggingFaceModel__pb2.fromPretrained_response.FromString,
                _registered_method=True)
        self.getBertModelForSequenceClassification = channel.unary_unary(
                '/mytransformers.HuggingFaceModelService/getBertModelForSequenceClassification',
                request_serializer=mytransformers_dot_HuggingFaceModel__pb2.fromPretrained_request.SerializeToString,
                response_deserializer=mytransformers_dot_HuggingFaceModel__pb2.fromPretrained_response.FromString,
                _registered_method=True)
        self.getRobertaModelForSequenceClassification = channel.unary_unary(
                '/mytransformers.HuggingFaceModelService/getRobertaModelForSequenceClassification',
                request_serializer=mytransformers_dot_HuggingFaceModel__pb2.fromPretrained_request.SerializeToString,
                response_deserializer=mytransformers_dot_HuggingFaceModel__pb2.fromPretrained_response.FromString,
                _registered_method=True)
        self.getT5ForConditionalGeneration = channel.unary_unary(
                '/mytransformers.HuggingFaceModelService/getT5ForConditionalGeneration',
                request_serializer=mytransformers_dot_HuggingFaceModel__pb2.fromPretrained_request.SerializeToString,
                response_deserializer=mytransformers_dot_HuggingFaceModel__pb2.fromPretrained_response.FromString,
                _registered_method=True)
        self.getAutoModelForSeq2SeqLM = channel.unary_unary(
                '/mytransformers.HuggingFaceModelService/getAutoModelForSeq2SeqLM',
                request_serializer=mytransformers_dot_HuggingFaceModel__pb2.fromPretrained_request.SerializeToString,
                response_deserializer=mytransformers_dot_HuggingFaceModel__pb2.fromPretrained_response.FromString,
                _registered_method=True)
        self.getXLNetForSequenceClassification = channel.unary_unary(
                '/mytransformers.HuggingFaceModelService/getXLNetForSequenceClassification',
                request_serializer=mytransformers_dot_HuggingFaceModel__pb2.fromPretrained_request.SerializeToString,
                response_deserializer=mytransformers_dot_HuggingFaceModel__pb2.fromPretrained_response.FromString,
                _registered_method=True)
        self.generate = channel.unary_unary(
                '/mytransformers.HuggingFaceModelService/generate',
                request_serializer=mytransformers_dot_HuggingFaceModel__pb2.generate_request.SerializeToString,
                response_deserializer=mytransformers_dot_HuggingFaceModel__pb2.generate_response.FromString,
                _registered_method=True)


class HuggingFaceModelServiceServicer(object):
    """/////////////////////////////////////////////////////////////////////////////////
    Service definition
    /////////////////////////////////////////////////////////////////////////////////
    """

    def getAutoModelForCausalML(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def getAutoModelForSequenceClassification(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def getBertModelForSequenceClassification(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def getRobertaModelForSequenceClassification(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def getT5ForConditionalGeneration(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def getAutoModelForSeq2SeqLM(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def getXLNetForSequenceClassification(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def generate(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_HuggingFaceModelServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'getAutoModelForCausalML': grpc.unary_unary_rpc_method_handler(
                    servicer.getAutoModelForCausalML,
                    request_deserializer=mytransformers_dot_HuggingFaceModel__pb2.fromPretrained_request.FromString,
                    response_serializer=mytransformers_dot_HuggingFaceModel__pb2.fromPretrained_response.SerializeToString,
            ),
            'getAutoModelForSequenceClassification': grpc.unary_unary_rpc_method_handler(
                    servicer.getAutoModelForSequenceClassification,
                    request_deserializer=mytransformers_dot_HuggingFaceModel__pb2.fromPretrained_request.FromString,
                    response_serializer=mytransformers_dot_HuggingFaceModel__pb2.fromPretrained_response.SerializeToString,
            ),
            'getBertModelForSequenceClassification': grpc.unary_unary_rpc_method_handler(
                    servicer.getBertModelForSequenceClassification,
                    request_deserializer=mytransformers_dot_HuggingFaceModel__pb2.fromPretrained_request.FromString,
                    response_serializer=mytransformers_dot_HuggingFaceModel__pb2.fromPretrained_response.SerializeToString,
            ),
            'getRobertaModelForSequenceClassification': grpc.unary_unary_rpc_method_handler(
                    servicer.getRobertaModelForSequenceClassification,
                    request_deserializer=mytransformers_dot_HuggingFaceModel__pb2.fromPretrained_request.FromString,
                    response_serializer=mytransformers_dot_HuggingFaceModel__pb2.fromPretrained_response.SerializeToString,
            ),
            'getT5ForConditionalGeneration': grpc.unary_unary_rpc_method_handler(
                    servicer.getT5ForConditionalGeneration,
                    request_deserializer=mytransformers_dot_HuggingFaceModel__pb2.fromPretrained_request.FromString,
                    response_serializer=mytransformers_dot_HuggingFaceModel__pb2.fromPretrained_response.SerializeToString,
            ),
            'getAutoModelForSeq2SeqLM': grpc.unary_unary_rpc_method_handler(
                    servicer.getAutoModelForSeq2SeqLM,
                    request_deserializer=mytransformers_dot_HuggingFaceModel__pb2.fromPretrained_request.FromString,
                    response_serializer=mytransformers_dot_HuggingFaceModel__pb2.fromPretrained_response.SerializeToString,
            ),
            'getXLNetForSequenceClassification': grpc.unary_unary_rpc_method_handler(
                    servicer.getXLNetForSequenceClassification,
                    request_deserializer=mytransformers_dot_HuggingFaceModel__pb2.fromPretrained_request.FromString,
                    response_serializer=mytransformers_dot_HuggingFaceModel__pb2.fromPretrained_response.SerializeToString,
            ),
            'generate': grpc.unary_unary_rpc_method_handler(
                    servicer.generate,
                    request_deserializer=mytransformers_dot_HuggingFaceModel__pb2.generate_request.FromString,
                    response_serializer=mytransformers_dot_HuggingFaceModel__pb2.generate_response.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'mytransformers.HuggingFaceModelService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('mytransformers.HuggingFaceModelService', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class HuggingFaceModelService(object):
    """/////////////////////////////////////////////////////////////////////////////////
    Service definition
    /////////////////////////////////////////////////////////////////////////////////
    """

    @staticmethod
    def getAutoModelForCausalML(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/mytransformers.HuggingFaceModelService/getAutoModelForCausalML',
            mytransformers_dot_HuggingFaceModel__pb2.fromPretrained_request.SerializeToString,
            mytransformers_dot_HuggingFaceModel__pb2.fromPretrained_response.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def getAutoModelForSequenceClassification(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/mytransformers.HuggingFaceModelService/getAutoModelForSequenceClassification',
            mytransformers_dot_HuggingFaceModel__pb2.fromPretrained_request.SerializeToString,
            mytransformers_dot_HuggingFaceModel__pb2.fromPretrained_response.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def getBertModelForSequenceClassification(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/mytransformers.HuggingFaceModelService/getBertModelForSequenceClassification',
            mytransformers_dot_HuggingFaceModel__pb2.fromPretrained_request.SerializeToString,
            mytransformers_dot_HuggingFaceModel__pb2.fromPretrained_response.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def getRobertaModelForSequenceClassification(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/mytransformers.HuggingFaceModelService/getRobertaModelForSequenceClassification',
            mytransformers_dot_HuggingFaceModel__pb2.fromPretrained_request.SerializeToString,
            mytransformers_dot_HuggingFaceModel__pb2.fromPretrained_response.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def getT5ForConditionalGeneration(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/mytransformers.HuggingFaceModelService/getT5ForConditionalGeneration',
            mytransformers_dot_HuggingFaceModel__pb2.fromPretrained_request.SerializeToString,
            mytransformers_dot_HuggingFaceModel__pb2.fromPretrained_response.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def getAutoModelForSeq2SeqLM(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/mytransformers.HuggingFaceModelService/getAutoModelForSeq2SeqLM',
            mytransformers_dot_HuggingFaceModel__pb2.fromPretrained_request.SerializeToString,
            mytransformers_dot_HuggingFaceModel__pb2.fromPretrained_response.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def getXLNetForSequenceClassification(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/mytransformers.HuggingFaceModelService/getXLNetForSequenceClassification',
            mytransformers_dot_HuggingFaceModel__pb2.fromPretrained_request.SerializeToString,
            mytransformers_dot_HuggingFaceModel__pb2.fromPretrained_response.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def generate(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/mytransformers.HuggingFaceModelService/generate',
            mytransformers_dot_HuggingFaceModel__pb2.generate_request.SerializeToString,
            mytransformers_dot_HuggingFaceModel__pb2.generate_response.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)



###############################################################################
# Copyright (c) 2024 MyTorch Systems Inc. All rights reserved.
###############################################################################

from torch.nn.Module import Module
from utils.logger import Logger
from proxies.mytransformers.HuggingFaceModel_proxy import HuggingFaceModelProxy
from torch.Tensor import Tensor
from utils.spinner import Spinner

class PreTrainedModel(Module):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.logger = Logger.get_logger()

    @classmethod
    def from_pretrained(cls, pretrained_model_name_or_path, *args, **kwargs):
        Logger.get_logger().info(f"Loading {cls.__name__} model `{pretrained_model_name_or_path}`")
        spinner = Spinner()
        spinner.start()

        try:
            # This method should be implemented by subclasses
            model_uuid = cls._get_model_uuid(pretrained_model_name_or_path, *args, **kwargs)
        finally:
            spinner.stop()

        model = cls(uuid=model_uuid)  # calls the class's __init__ method
        Logger.get_logger().info(f"...model has been loaded")
        return model

    @classmethod
    def _get_model_uuid(cls, pretrained_model_name_or_path, *args, **kwargs):
        # This method should be implemented by subclasses
        raise NotImplementedError("Subclasses must implement the _get_model_uuid method.")


    def generate(self, **kwargs):
        tensor_uuid = HuggingFaceModelProxy().generate(self.uuid, **kwargs)
        return Tensor(uuid=tensor_uuid)


# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: mytorchvision/transforms/transforms.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from gRPC_impl import shared_msg_types_pb2 as shared__msg__types__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n)mytorchvision/transforms/transforms.proto\x12\rmytorchvision\x1a\x16shared_msg_types.proto\"\x1d\n\rGrpcTransform\x12\x0c\n\x04uuid\x18\x01 \x01(\t\"0\n CreateCenterCropTransformRequest\x12\x0c\n\x04size\x18\x01 \x01(\x05\",\n\x1c\x43reateResizeTransformRequest\x12\x0c\n\x04size\x18\x01 \x01(\x05\"<\n\x1f\x43reateNormalizeTransformRequest\x12\x0c\n\x04mean\x18\x01 \x03(\x02\x12\x0b\n\x03std\x18\x02 \x03(\x02\"A\n\x1d\x43reateComposeTransformRequest\x12 \n\x18\x63omponent_transform_uuid\x18\x01 \x03(\t2\xfb\x03\n\x11TransformsService\x12j\n\x19\x43reateCenterCropTransform\x12/.mytorchvision.CreateCenterCropTransformRequest\x1a\x1c.mytorchvision.GrpcTransform\x12\x62\n\x15\x43reateResizeTransform\x12+.mytorchvision.CreateResizeTransformRequest\x1a\x1c.mytorchvision.GrpcTransform\x12\x46\n\x17\x43reateToTensorTransform\x12\r.shared.Empty\x1a\x1c.mytorchvision.GrpcTransform\x12h\n\x18\x43reateNormalizeTransform\x12..mytorchvision.CreateNormalizeTransformRequest\x1a\x1c.mytorchvision.GrpcTransform\x12\x64\n\x16\x43reateComposeTransform\x12,.mytorchvision.CreateComposeTransformRequest\x1a\x1c.mytorchvision.GrpcTransformb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'mytorchvision.transforms.transforms_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_GRPCTRANSFORM']._serialized_start=84
  _globals['_GRPCTRANSFORM']._serialized_end=113
  _globals['_CREATECENTERCROPTRANSFORMREQUEST']._serialized_start=115
  _globals['_CREATECENTERCROPTRANSFORMREQUEST']._serialized_end=163
  _globals['_CREATERESIZETRANSFORMREQUEST']._serialized_start=165
  _globals['_CREATERESIZETRANSFORMREQUEST']._serialized_end=209
  _globals['_CREATENORMALIZETRANSFORMREQUEST']._serialized_start=211
  _globals['_CREATENORMALIZETRANSFORMREQUEST']._serialized_end=271
  _globals['_CREATECOMPOSETRANSFORMREQUEST']._serialized_start=273
  _globals['_CREATECOMPOSETRANSFORMREQUEST']._serialized_end=338
  _globals['_TRANSFORMSSERVICE']._serialized_start=341
  _globals['_TRANSFORMSSERVICE']._serialized_end=848
# @@protoc_insertion_point(module_scope)



###############################################################################
# Copyright (c) 2024 MyTorch Systems Inc. All rights reserved.
###############################################################################

'''
An AutoModel is a generalized factory class that can be used to instantiate a model from a pretrained model.
In almost all cases, a subclass of AutoModel should be used instead of AutoModel itself. Subclasses of AutoModel
are specialized for specific types of models, such as AutoModelForCausalLM for causal language modeling or
AutoModelForSequenceClassification for sequence classification. These subclasses provide additional functionality
and are more user-friendly than AutoModel.
'''

from transformers.PreTrainedModel import PreTrainedModel

class AutoModel(PreTrainedModel):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: mytorch/scaffolding/scaffolding.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from gRPC_impl import shared_msg_types_pb2 as shared__msg__types__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1bmytorch/scaffolding/scaffolding.proto\x12\x07mytorch\x1a\x16shared_msg_types.proto\"*\n\x0cKeyValuePair\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t\"<\n\nClientInfo\x12.\n\x0fkey_value_pairs\x18\x02 \x03(\x0b\x32\x15.mytorch.KeyValuePair2\x8c\x01\n\x0eMyTorchService\x12\x44\n\x1cinitialize_client_connection\x12\x13.mytorch.ClientInfo\x1a\r.shared.Empty\"\x00\x12\x34\n\x0cget_gpu_info\x12\r.shared.Empty\x1a\x13.shared.StringValue\"\x00\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'mytorch.scaffolding.scaffolding_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_KEYVALUEPAIR']._serialized_start=64
  _globals['_KEYVALUEPAIR']._serialized_end=106
  _globals['_CLIENTINFO']._serialized_start=108
  _globals['_CLIENTINFO']._serialized_end=168
  _globals['_MYTORCHSERVICE']._serialized_start=171
  _globals['_MYTORCHSERVICE']._serialized_end=311
# @@protoc_insertion_point(module_scope)



from gRPC_impl import shared_msg_types_pb2 as _shared_msg_types_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class ARangeRequest(_message.Message):
    __slots__ = ("start", "end", "step")
    START_FIELD_NUMBER: _ClassVar[int]
    END_FIELD_NUMBER: _ClassVar[int]
    STEP_FIELD_NUMBER: _ClassVar[int]
    start: float
    end: float
    step: float
    def __init__(self, start: _Optional[float] = ..., end: _Optional[float] = ..., step: _Optional[float] = ...) -> None: ...

class ArgMaxRequest(_message.Message):
    __slots__ = ("tensor_uuid", "dim", "keepdim")
    TENSOR_UUID_FIELD_NUMBER: _ClassVar[int]
    DIM_FIELD_NUMBER: _ClassVar[int]
    KEEPDIM_FIELD_NUMBER: _ClassVar[int]
    tensor_uuid: str
    dim: int
    keepdim: bool
    def __init__(self, tensor_uuid: _Optional[str] = ..., dim: _Optional[int] = ..., keepdim: bool = ...) -> None: ...

class AllCloseRequest(_message.Message):
    __slots__ = ("tensor1_uuid", "tensor2_uuid", "rtol", "atol", "equal_nan")
    TENSOR1_UUID_FIELD_NUMBER: _ClassVar[int]
    TENSOR2_UUID_FIELD_NUMBER: _ClassVar[int]
    RTOL_FIELD_NUMBER: _ClassVar[int]
    ATOL_FIELD_NUMBER: _ClassVar[int]
    EQUAL_NAN_FIELD_NUMBER: _ClassVar[int]
    tensor1_uuid: str
    tensor2_uuid: str
    rtol: float
    atol: float
    equal_nan: bool
    def __init__(self, tensor1_uuid: _Optional[str] = ..., tensor2_uuid: _Optional[str] = ..., rtol: _Optional[float] = ..., atol: _Optional[float] = ..., equal_nan: bool = ...) -> None: ...

class RepeatInterleaveRequest(_message.Message):
    __slots__ = ("tensor_uuid", "repeats", "dim")
    TENSOR_UUID_FIELD_NUMBER: _ClassVar[int]
    REPEATS_FIELD_NUMBER: _ClassVar[int]
    DIM_FIELD_NUMBER: _ClassVar[int]
    tensor_uuid: str
    repeats: int
    dim: int
    def __init__(self, tensor_uuid: _Optional[str] = ..., repeats: _Optional[int] = ..., dim: _Optional[int] = ...) -> None: ...

class SaveModelRequest(_message.Message):
    __slots__ = ("filename", "named_tensor_uuids")
    class NamedTensorUuidsEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
    FILENAME_FIELD_NUMBER: _ClassVar[int]
    NAMED_TENSOR_UUIDS_FIELD_NUMBER: _ClassVar[int]
    filename: str
    named_tensor_uuids: _containers.ScalarMap[str, str]
    def __init__(self, filename: _Optional[str] = ..., named_tensor_uuids: _Optional[_Mapping[str, str]] = ...) -> None: ...

class LoadModelResponse(_message.Message):
    __slots__ = ("tensors",)
    class TensorsEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: _shared_msg_types_pb2.GrpcTensor
        def __init__(self, key: _Optional[str] = ..., value: _Optional[_Union[_shared_msg_types_pb2.GrpcTensor, _Mapping]] = ...) -> None: ...
    TENSORS_FIELD_NUMBER: _ClassVar[int]
    tensors: _containers.MessageMap[str, _shared_msg_types_pb2.GrpcTensor]
    def __init__(self, tensors: _Optional[_Mapping[str, _shared_msg_types_pb2.GrpcTensor]] = ...) -> None: ...



from gRPC_impl import shared_msg_types_pb2 as _shared_msg_types_pb2
from gRPC_impl.mytorch.nn import nn_msg_types_pb2 as _nn_msg_types_pb2
from google.protobuf import descriptor as _descriptor
from typing import ClassVar as _ClassVar

DESCRIPTOR: _descriptor.FileDescriptor



###############################################################################
# Copyright (c) 2024 MyTorch Systems Inc. All rights reserved.
###############################################################################

"""
This proxy provides a proxy for the Module gRPC service.
It allows the client to call specified torch.nn.Module operations on the server via gRPC.
"""

from connection_utils.server_connection import ServerConnection, wrap_with_error_handler
from gRPC_impl.mytorch.nn import module_pb2_grpc, nn_msg_types_pb2, module_pb2
from gRPC_impl import shared_msg_types_pb2
from utils.data_transform_utils import deserialize_state_dict
from utils.logger import Logger
from torch.Tensor import Tensor
from torch.nn.ParametersGenerator import ParametersGenerator
from proxies.mytorch.scaffolding.scaffolding_proxy import ScaffoldingProxy
from proxies.base_proxy import BaseProxy

import numpy as np

class ModuleProxy( BaseProxy ):
    # if uuid is None, it means that the module is not yet created and
    # one will be created when initializing the proxy
    def __init__(self):
        self.channel = ServerConnection.get_active_connection()
        self.module_stub = module_pb2_grpc.ModuleServiceStub(self.channel)
        self.logger = Logger.get_logger()
        self.uuid = None

    @wrap_with_error_handler
    def create_module_on_server(self):
        # Notify the server to create a real torch.nn.Module
        return_val = self.module_stub.CreateModuleOnServer(shared_msg_types_pb2.Empty())
        self.uuid = return_val.uuid
        self.logger.info("ModuleProxy: Created proxy with ID: " + return_val.uuid)
        return return_val.uuid

    @wrap_with_error_handler
    def state_dict(self):
        self.logger.debug(f"Getting state_dict for proxy with ID `{self.uuid}`")
        # Request the state_dict from the server
        request = shared_msg_types_pb2.UUID(uuid=self.uuid)
        stream = self.module_stub.GetStateDict(request)

        state_dict = {}
        for named_tensor in stream:
            tensor_proxy = Tensor(
                uuid=named_tensor.tensor.uuid,
                shape=list(named_tensor.tensor.shape),
                dtype=named_tensor.tensor.dtype,
            )
            state_dict[named_tensor.name] = tensor_proxy

        return state_dict

    @wrap_with_error_handler
    def load_state_dict(self, state_dict: dict) -> None:
        self.logger.debug(f"Uploading state_dict to module `{self.uuid}`")

        def chunk_generator():
            for name, tensor in state_dict.items():
                if not isinstance(tensor, Tensor):
                    raise TypeError(f"Expected Tensor proxy for key '{name}', got {type(tensor)}")

                grpc_tensor = shared_msg_types_pb2.GrpcTensor(
                    uuid=tensor.uuid,
                    shape=tensor.shape,
                    dtype=tensor.dtype,
                )

                yield shared_msg_types_pb2.NamedTensor(
                    name=name,
                    tensor=grpc_tensor
                )

        # Server will use self.uuid to apply state_dict
        
        metadata = (("module-uuid", self.uuid),)
        response = self.module_stub.LoadStateDict(chunk_generator(), metadata=metadata)

        if not response.value:
            raise RuntimeError("Failed to load state_dict on the server.")


    @wrap_with_error_handler
    def forward(self, input_data) -> Tensor:
        # for some reason, right now the train_test_split function
        # is returning a list of numpy arrays, so convert it here
        # TODO: fix this in the train_test_split function
        if isinstance(input_data, np.ndarray):
            tensor_uuid = MyTorchProxy().from_numpy(input_data).uuid

        # if the input_data is a PyTorch tensor, send it to the server
        # and get the UUID of the tensor on the server
        # elif isinstance(input_data, torch.Tensor):
        #     tensor_uuid = MyTorchProxy().send_pytorch_tensor_to_server(input_data).uuid

        # this is a MyTorch tensor
        else:
            tensor_uuid = input_data.uuid

        request = nn_msg_types_pb2.ForwardPassRequest()
        request.module_uuid = self.uuid
        request.tensor_uuid = tensor_uuid
        response: shared_msg_types_pb2.GrpcTensor = self.module_stub.Forward(request)
        return Tensor(response.uuid, response.shape, response.dtype)

    @wrap_with_error_handler
    def eval(self):
        request = shared_msg_types_pb2.UUID(uuid=self.uuid)
        self.module_stub.Eval(request)

    @wrap_with_error_handler
    def to_device(self, device: str) -> None:
        request = module_pb2.ModuleToDeviceRequest(uuid=self.uuid, device=device)
        response = self.module_stub.ToDevice(request)

    @wrap_with_error_handler
    def parameters(self):
        """
        Uses generic_call instead of gRPC GetParameters.
        Calls model.parameters() using call_type="method".
        Returns a list of Tensor proxy objects.
        """
        results = self.generic_call("torch.nn", "parameters", self.uuid, call_type="method")

        # The response should be a list of (uuid, shape, dtype) tuples
        if not isinstance(results, list):
            raise RuntimeError("Expected list of parameters from generic_call")

        tensors = []
        for entry in results:
            if not isinstance(entry, tuple) or len(entry) != 3:
                raise ValueError(f"Invalid tensor response format: {entry}")
            uuid, shape, dtype = entry
            tensors.append(Tensor(uuid=uuid, shape=shape, dtype=dtype))

        return tensors

    @wrap_with_error_handler
    def half(self):
        request = module_pb2.half_request(uuid=self.uuid)
        self.module_stub.half(request)

    ### KEY METHOD: When objects go out of scope, this is called to clean up server memory.
    # See also TensorProxy.delete
    @wrap_with_error_handler
    def delete(self):
        request = shared_msg_types_pb2.UUID(uuid=self.uuid)
        self.module_stub.delete(request)

    @wrap_with_error_handler
    def train(self, mode=True):
        # Torch uses .train(True) or .train(False)
        return self.generic_call("torch.nn", "train", self.uuid, mode, call_type="method")



###############################################################################
# Copyright (c) 2024 MyTorch Systems Inc. All rights reserved.
###############################################################################
from .Module import Module
from .Sequential import Sequential
from .Linear import Linear
from .ReLU import ReLU
from .Flatten import Flatten
from .CrossEntropyLoss import CrossEntropyLoss
from .LSTM import LSTM
from .Sigmoid import Sigmoid
from .MSELoss import MSELoss

from . import functional

from .functional import softmax

__all__ = [
     'Module',
     'Sequential',
     'Linear',
     'ReLU',
     'Flatten',
     'CrossEntropyLoss',
     'functional',
     'softmax',
     'LSTM'
 ]


# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: mytorchvision/datasets/datasets.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n%mytorchvision/datasets/datasets.proto\x12\rmytorchvision\"U\n\x0fGrpcImageFolder\x12\x0c\n\x04uuid\x18\x01 \x01(\t\x12\x13\n\x0bimage_uuids\x18\x02 \x03(\t\x12\x0e\n\x06labels\x18\x03 \x03(\x05\x12\x0f\n\x07\x63lasses\x18\x04 \x03(\t\"D\n\x0bGrpcCIFAR10\x12\x0c\n\x04uuid\x18\x01 \x01(\t\x12\x16\n\x0e\x64\x61taset_length\x18\x02 \x01(\x05\x12\x0f\n\x07\x63lasses\x18\x03 \x03(\t\"M\n\x18\x43reateImageFolderRequest\x12\x19\n\x11relative_datapath\x18\x01 \x01(\t\x12\x16\n\x0etransform_uuid\x18\x02 \x01(\t\"]\n\x14\x43reateCIFAR10Request\x12\x0c\n\x04root\x18\x01 \x01(\t\x12\r\n\x05train\x18\x02 \x01(\x08\x12\x10\n\x08\x64ownload\x18\x03 \x01(\x08\x12\x16\n\x0etransform_uuid\x18\x04 \x01(\t2\xd1\x01\n\x0f\x44\x61tasetsService\x12\x64\n\x19\x43reateImageFolderOnServer\x12\'.mytorchvision.CreateImageFolderRequest\x1a\x1e.mytorchvision.GrpcImageFolder\x12X\n\x15\x43reateCIFAR10onServer\x12#.mytorchvision.CreateCIFAR10Request\x1a\x1a.mytorchvision.GrpcCIFAR10b\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'mytorchvision.datasets.datasets_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_GRPCIMAGEFOLDER']._serialized_start=56
  _globals['_GRPCIMAGEFOLDER']._serialized_end=141
  _globals['_GRPCCIFAR10']._serialized_start=143
  _globals['_GRPCCIFAR10']._serialized_end=211
  _globals['_CREATEIMAGEFOLDERREQUEST']._serialized_start=213
  _globals['_CREATEIMAGEFOLDERREQUEST']._serialized_end=290
  _globals['_CREATECIFAR10REQUEST']._serialized_start=292
  _globals['_CREATECIFAR10REQUEST']._serialized_end=385
  _globals['_DATASETSSERVICE']._serialized_start=388
  _globals['_DATASETSSERVICE']._serialized_end=597
# @@protoc_insertion_point(module_scope)



###############################################################################
# Copyright (c) 2024 MyTorch Systems Inc. All rights reserved.
###############################################################################

from proxies.base_proxy import BaseProxy
from torch.Tensor import Tensor
from connection_utils.server_connection import ServerConnection

class MSELoss:
    def __init__(self, reduction="mean"):
        self.reduction = reduction
        self.proxy = BaseProxy()
        self.proxy.channel = ServerConnection.get_active_connection()
        self.uuid = self.proxy.generic_call(
            "torch.nn", "MSELoss",
            call_type="constructor",
            kwargs={"reduction": self.reduction}
        )
    def __call__(self, input: Tensor, target: Tensor):
        uuid, shape, dtype = self.proxy.generic_call(
            "torch.nn", "forward",
            self.uuid, input.uuid, target.uuid,
            call_type="forward"
        )
        return Tensor(uuid, shape, dtype)



from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Optional as _Optional

DESCRIPTOR: _descriptor.FileDescriptor

class TraceRequest(_message.Message):
    __slots__ = ("model_uuid", "tensor_uuid_tuple")
    MODEL_UUID_FIELD_NUMBER: _ClassVar[int]
    TENSOR_UUID_TUPLE_FIELD_NUMBER: _ClassVar[int]
    model_uuid: str
    tensor_uuid_tuple: _containers.RepeatedScalarFieldContainer[str]
    def __init__(self, model_uuid: _Optional[str] = ..., tensor_uuid_tuple: _Optional[_Iterable[str]] = ...) -> None: ...

class TraceResponse(_message.Message):
    __slots__ = ("model_uuid",)
    MODEL_UUID_FIELD_NUMBER: _ClassVar[int]
    model_uuid: str
    def __init__(self, model_uuid: _Optional[str] = ...) -> None: ...



###############################################################################
# Copyright (c) 2024 MyTorch Systems Inc. All rights reserved.
###############################################################################

"""
This proxy provides utility methods for deserializing tensors and state_dicts.

"""

import numpy as np
from gRPC_impl import shared_msg_types_pb2

def convert_str_to_numpy_dtype(dtype_str) -> np.dtype:
    """Converts a dtype string from PyTorch format to NumPy dtype."""
    dtype_map = {
        'torch.float32': np.float32,
        'torch.float64': np.float64,
        'torch.float16': np.float16,
        'torch.int32': np.int32,
        'torch.int64': np.int64,
        'torch.int16': np.int16,
        'torch.int8': np.int8,
        'torch.uint8': np.uint8,
    }
    return dtype_map.get(dtype_str, np.float32)  # Default to np.float32 if not found

def deserialize_tensor(tensor_data, tensor_shape, dtype_str) -> np.ndarray:
    """Deserializes tensor data (bytes) into a NumPy array."""
    dtype: np.dtype = convert_str_to_numpy_dtype(dtype_str)  # Use the utility method to convert dtype
    shape_tuple = tuple(tensor_shape) # Convert the RepeatedScalarContainer of the gRPC data to a tuple
    return np.frombuffer(tensor_data, dtype=dtype).reshape(shape_tuple)

def deserialize_state_dict(serialized_state_dict) -> dict:
    """Deserializes a serialized state_dict (protobuf message) into a dictionary of NumPy arrays."""
    state_dict_np = {}
    for param_name, serialized_tensor in serialized_state_dict.entries.items():
        tensor_np = deserialize_tensor(serialized_tensor.data,
                                       serialized_tensor.shape,
                                       dtype_str=serialized_tensor.dtype)
        state_dict_np[param_name] = tensor_np
    return state_dict_np

def serialize_numpy_array(np_array: np.ndarray) -> shared_msg_types_pb2.SerializedNumpyArray:
    array_data = np_array.tobytes()
    array_shape = list(np_array.shape)
    array_dtype = str(np_array.dtype)
    return shared_msg_types_pb2.SerializedNumpyArray(
            data=array_data,
            shape=array_shape,
            dtype=array_dtype
        )

def convert_python_to_grpc_value(value):
    grpc_value = shared_msg_types_pb2.Value()
    if isinstance(value, bool):
        grpc_value.bool_value = value
    elif isinstance(value, int):
        grpc_value.int_value = value
    elif isinstance(value, float):
        grpc_value.float_value = value
    elif isinstance(value, str):
        grpc_value.string_value = value
    else:
        grpc_value.string_value = str(value)
    return grpc_value

def convert_python_to_grpc_kwargs(kwargs):
    return {k: convert_python_to_grpc_value(v) for k, v in kwargs.items()}

def convert_grpc_value_to_python(value):
    # If it's already a Python primitive type, return it directly
    if isinstance(value, (str, int, float, bool)):
        return value
        
    # Otherwise handle gRPC Value type
    which_oneof = value.WhichOneof('kind')
    if which_oneof == 'string_value':
        return value.string_value
    elif which_oneof == 'int_value':
        return value.int_value
    elif which_oneof == 'float_value':
        return value.float_value
    elif which_oneof == 'bool_value':
        return value.bool_value
    return None

def convert_grpc_kwargs_to_python(kwargs):
    # Convert the entire kwargs dictionary
    python_kwargs = {
        key: convert_grpc_value_to_python(value) 
        for key, value in kwargs.items()
    }
    return python_kwargs



###############################################################################
# Copyright (c) 2024 MyTorch Systems Inc. All rights reserved.
###############################################################################
from .Transformer import Transformer
from proxies.mytorchvision.transforms.NormalizeProxy import NormalizeProxy

class Normalize (Transformer):
    def __init__(self, mean: list, std: list):
        super().__init__()
        self.mean = mean
        self.std = std
        self.proxy = NormalizeProxy(mean, std)
        self.uuid = self.proxy.createTransformOnServer()



###############################################################################
# Copyright (c) 2024 MyTorch Systems Inc. All rights reserved.
###############################################################################

from proxies.mytorch.scaffolding.data_mgmt_proxy import DataMgmtProxy

def folder_exists_on_server(folder_path: str) -> bool:
    return DataMgmtProxy().folder_exists_on_server(folder_path)

def upload_folder_to_server(folder_path: str) -> int:
    return DataMgmtProxy().upload_folder(folder_path)


###############################################################################
# Copyright (c) 2024 MyTorch Systems Inc. All rights reserved.
###############################################################################

from proxies.mytorch.tensor_proxy import TensorProxy
from utils.data_transform_utils import convert_str_to_numpy_dtype
import numpy as np
from torch.device import device

class Tensor:
    def __init__(self, uuid: str, shape = None, dtype: str = None):
        self.proxy = TensorProxy(uuid)
        self.uuid = uuid

        # if shape not passed in, it will be fetched from the server
        # the first time it is needed
        self._shape = tuple(shape) if shape is not None else None

        # if dtype not passed in, it will be fetched from the server
        # the first time it is needed
        self._dtype = dtype

        # lazy loading of data
        self._retrieved_data = None

    @property
    def shape(self):
        # if shape not passed in, it will be fetched from the server
        # the first time it is needed
        if self._shape is None:
            # Fetch the shape from the server if it hasn't been fetched before
            self._shape = self.proxy.get_shape()
        return self._shape

    @property
    def dtype(self):
        # if dtype not passed in, it will be fetched from the server
        # the first time it is needed
        if self._dtype is None:
            # Fetch the shape from the server if it hasn't been fetched before
            self._dtype = self.proxy.get_dtype()
        return self._dtype

    @property
    def numpy_dtype(self):
       return convert_str_to_numpy_dtype(self.dtype)

    @property
    def retrieved_data(self) -> np.ndarray:
        if self._retrieved_data is None:
            # Assuming TensorProxy.get_data() returns a numpy.ndarray
            self._retrieved_data = self.proxy.get_data()
        return self._retrieved_data

    def size(self, dim = None):
        if dim is None:
            return self.shape
        return self.shape[dim]

    def dim(self):
        return len( self.shape)
    
    def __str__(self):
        return str(self.retrieved_data)

    def __len__(self):
        # Fetch the data shape and return the size of the first dimension
        return self.shape[0]

    def backward(self):
        self.proxy.backward()

    def item(self) -> float:
        return self.proxy.item()

    def float(self):
        uuid, shape, dtype = self.proxy.float()
        return Tensor(uuid, shape, dtype)

    def long(self):
        uuid, shape, dtype = self.proxy.long()
        return Tensor(uuid, shape, dtype)

    def numpy(self):
        return self.retrieved_data

    def cuda(self):
        uuid, shape, dtype = self.proxy.toCuda()
        return Tensor(uuid, shape, dtype)

    def cpu(self):
        uuid, shape, dtype = self.proxy.toCpu()
        return Tensor(uuid, shape, dtype)

    def sum(self, operand):
        uuid, shape, dtype = self.proxy.sum(operand)
        return Tensor(uuid, shape, dtype)
    
    def sub(self, operand):
        #return MyTorchProxy().generic_call("Tensor", "sub", self.uuid, operand)
        uuid, shape, dtype = self.proxy.sub(operand)
        return Tensor(uuid, shape, dtype)

            
    def to(self, device):
        # Convert device object to string if it's not already a string
        device_str = str(device) if not isinstance(device, str) else device
        
        if device_str == "cuda":
            return self.cuda()
        elif device_str == "cpu":
            return self.cpu()
        else:
            raise ValueError(f"Device {device_str} not recognized")

    def reshape(self, shape: tuple):
        uuid, shape, dtype = self.proxy.reshape(shape)
        return Tensor(uuid, shape, dtype)

    def unsqueeze(self, dim):
        uuid, shape, dtype = self.proxy.unsqueeze(dim)
        return Tensor(uuid, shape, dtype)

    def squeeze(self, dim):
        uuid, shape, dtype = self.proxy.squeeze(dim)
        return Tensor(uuid, shape, dtype)

    def view(self, batch_size: int, input_size: int):
        uuid, shape, dtype = self.proxy.view(batch_size, input_size )
        return Tensor(uuid, shape, dtype)

    # transpose
    def t(self):
        uuid, shape, dtype = self.proxy.t()
        return Tensor(uuid, shape, dtype)

    def chunk(self, chunks: int, dim: int = 0):
        """
        Splits the tensor into 'chunks' number of smaller tensors along dimension 'dim'.
        Returns a tuple of new mytorch.Tensor objects.
        """
        # Call the proxy, which returns the uuids, shapes, and dtypes of each chunk
        chunk_info = self.proxy.chunk(chunks, dim)

        # Convert each returned (uuid, shape, dtype) triple into a mytorch.Tensor
        chunk_tensors = []
        for (uuid, shape, dtype) in chunk_info:
            chunk_tensors.append(Tensor(uuid, shape, dtype))

        # PyTorch's chunk returns a tuple, so we'll do the same
        return tuple(chunk_tensors)


    #####################################################################################
    # These methods are needed for the tensor to be treated like a
    # numpy array in some cases, such as with sckit-learn's train_test_split
    #####################################################################################

    # return an actual NumPy array
    def __array__(self, dtype=None):
        return np.asarray(self.retrieved_data, dtype=dtype or self.numpy_dtype)

    # for indexing and slicing
    def __getitem__(self, idx):
        # indexing, e.g. tensor[0]
        if isinstance(idx, int):
            uuid, shape, dtype = self.proxy.index(idx)
            return Tensor(uuid, shape, dtype)

        # slicing, e.g. tensor[0:2,3:,:5]
        else:
            uuid, shape, dtype = self.proxy.slice(idx)
            return Tensor(uuid, shape, dtype)

    # support iteration
    def __iter__(self):
        return iter(self.retrieved_data)

    def __add__(self, other):
        uuid, shape, dtype = self.proxy.add(other.uuid)
        return Tensor( uuid, shape, dtype)

    def __radd__(self, other):
        """Handle case when tensor is on right side of addition."""
        return self.__add__(other)

    def __sub__(self, other):
        uuid, shape, dtype = self.proxy.sub(other.uuid)
        return Tensor( uuid, shape, dtype)

    #def __mul__(self, other):
    #    uuid, shape, dtype = self.proxy.mul(other.uuid)
    #    return Tensor( uuid, shape, dtype)

    def __mul__(self, other):
        operand = other.uuid if isinstance(other, Tensor) else other
        uuid, shape, dtype = self.proxy.mul(operand)
        return Tensor(uuid, shape, dtype)

    def __pow__(self, operand):
        uuid, shape, dtype = self.proxy.pow(operand)
        return Tensor( uuid, shape, dtype)
    
    def __rmul__(self, other):
        return self.__mul__(other)

    # allows for comparison with other tensors
    def __eq__(self, other):
        uuid, shape, dtype =self.proxy.equal(other.uuid)
        return Tensor(uuid, shape, dtype)

    def __del__(self):
        self.proxy.delete()
        pass



###############################################################################
# Copyright (c) 2024-2025 MyTorch Systems Inc. All rights reserved.
###############################################################################

from proxies.base_proxy import BaseProxy
from utils.logger import Logger
from connection_utils.server_connection import ServerConnection, wrap_with_error_handler
from proxies.mytorch.optim.optimizer_proxy import OptimizerProxy


class AdamProxy(OptimizerProxy, BaseProxy):  # Inherit from both
    def __init__(self):
        super().__init__()

        self.logger = Logger.get_logger()
        self.channel = ServerConnection.get_active_connection()
        self.uuid = None  # assigned when optimizer is registered on server

    @wrap_with_error_handler
    def zero_grad(self):
        return self.generic_call("torch.optim", "zero_grad", self.uuid, call_type="method")

    @wrap_with_error_handler
    def step(self):
        return self.generic_call("torch.optim", "step", self.uuid, call_type="method")
    



###############################################################################
# Copyright (c) 2024 MyTorch Systems Inc. All rights reserved.
###############################################################################

'''
This is a factory class that can be used to instantiate a Bert model for sequence classification from a pretrained model.
'''

from transformers.PreTrainedModel import PreTrainedModel
from proxies.mytransformers.HuggingFaceModel_proxy import HuggingFaceModelProxy

class T5ForConditionalGeneration(PreTrainedModel):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @classmethod
    def _get_model_uuid(cls, pretrained_model_name_or_path, *args, **kwargs):
        return HuggingFaceModelProxy().getT5ForConditionalGeneration(pretrained_model_name_or_path, *args, **kwargs)
